# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Spyder path manager."""

# Standard library imports
from collections import OrderedDict
import os
import os.path as osp
import sys

# Third party imports
from qtpy import PYQT5, PYQT6
from qtpy.compat import getexistingdirectory
from qtpy.QtCore import QSize, Qt, Signal, Slot
from qtpy.QtGui import QFontMetrics
from qtpy.QtWidgets import (QDialog, QDialogButtonBox, QHBoxLayout,
                            QListWidget, QListWidgetItem, QMessageBox,
                            QVBoxLayout, QLabel)

# Local imports
from spyder.api.widgets.dialogs import SpyderDialogButtonBox
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.config.base import _
from spyder.plugins.pythonpath.utils import check_path, get_system_pythonpath
from spyder.utils.environ import get_user_env, set_user_env
from spyder.utils.misc import getcwd_or_home
from spyder.utils.stylesheet import (
    AppStyle,
    MAC,
    PANES_TOOLBAR_STYLESHEET,
    WIN
)


class PathManagerToolbuttons:
    MoveTop = 'move_top'
    MoveUp = 'move_up'
    MoveDown = 'move_down'
    MoveToBottom = 'move_to_bottom'
    AddPath = 'add_path'
    RemovePath = 'remove_path'
    ExportPaths = 'export_paths'


class PathManager(QDialog, SpyderWidgetMixin):
    """Path manager dialog."""

    redirect_stdio = Signal(bool)
    sig_path_changed = Signal(object)

    # This is required for our tests
    CONF_SECTION = 'pythonpath_manager'

    def __init__(self, parent, path=None, project_path=None,
                 not_active_path=None, sync=True):
        """Path manager dialog."""
        if PYQT5 or PYQT6:
            super().__init__(parent, class_parent=parent)
        else:
            QDialog.__init__(self, parent)
            SpyderWidgetMixin.__init__(self, class_parent=parent)

        assert isinstance(path, (tuple, type(None)))

        # Style
        # NOTE: This needs to be here so all buttons are styled correctly
        self.setStyleSheet(self._stylesheet)

        self.path = path or ()
        self.project_path = project_path or ()
        self.not_active_path = not_active_path or ()

        self.last_path = getcwd_or_home()
        self.original_path_dict = None
        self.system_path = ()
        self.user_path = []

        # This is necessary to run our tests
        if self.path:
            self.update_paths(system_path=get_system_pythonpath())

        # Widgets
        self.add_button = None
        self.remove_button = None
        self.movetop_button = None
        self.moveup_button = None
        self.movedown_button = None
        self.movebottom_button = None
        self.export_button = None
        self.user_header = None
        self.project_header = None
        self.system_header = None
        self.headers = []
        self.selection_widgets = []
        self.right_buttons = self._setup_right_toolbar()
        self.listwidget = QListWidget(self)
        self.bbox = SpyderDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        self.button_ok = self.bbox.button(QDialogButtonBox.Ok)

        # Widget setup
        self.setWindowTitle(_("PYTHONPATH manager"))
        self.setWindowIcon(self.create_icon('pythonpath'))
        self.resize(500, 400)
        self.export_button.setVisible(os.name == 'nt' and sync)

        # Description
        description = QLabel(
            _("The paths listed below will be passed to the IPython console "
              "and to the Editor as additional locations to search for Python "
              "modules.")
        )
        description.setWordWrap(True)

        # Buttons layout
        buttons_layout = QVBoxLayout()
        self._add_buttons_to_layout(self.right_buttons, buttons_layout)
        buttons_layout.addStretch(1)

        # Middle layout
        middle_layout = QHBoxLayout()
        middle_layout.setContentsMargins(4 if WIN else 5, 0, 0, 0)
        middle_layout.addWidget(self.listwidget)
        middle_layout.addLayout(buttons_layout)

        # Widget layout
        layout = QVBoxLayout()
        layout.addWidget(description)
        layout.addSpacing(2 * AppStyle.MarginSize)
        layout.addLayout(middle_layout)
        layout.addSpacing((-1 if MAC else 2) * AppStyle.MarginSize)
        layout.addWidget(self.bbox)
        self.setLayout(layout)

        # Signals
        self.listwidget.currentRowChanged.connect(lambda x: self.refresh())
        self.listwidget.itemChanged.connect(lambda x: self.refresh())
        self.bbox.accepted.connect(self.accept)
        self.bbox.rejected.connect(self.reject)

        # Setup
        self.setup()

    # ---- Private methods
    # -------------------------------------------------------------------------
    def _add_buttons_to_layout(self, widgets, layout):
        """Helper to add buttons to its layout."""
        for widget in widgets:
            layout.addWidget(widget)

    def _setup_right_toolbar(self):
        """Create top toolbar and actions."""
        self.movetop_button = self.create_toolbutton(
            PathManagerToolbuttons.MoveTop,
            text=_("Move path to the top"),
            icon=self.create_icon('2uparrow'),
            triggered=lambda: self.move_to(absolute=0))
        self.moveup_button = self.create_toolbutton(
            PathManagerToolbuttons.MoveUp,
            tip=_("Move path up"),
            icon=self.create_icon('1uparrow'),
            triggered=lambda: self.move_to(relative=-1))
        self.movedown_button = self.create_toolbutton(
            PathManagerToolbuttons.MoveDown,
            tip=_("Move path down"),
            icon=self.create_icon('1downarrow'),
            triggered=lambda: self.move_to(relative=1))
        self.movebottom_button = self.create_toolbutton(
            PathManagerToolbuttons.MoveToBottom,
            text=_("Move path to the bottom"),
            icon=self.create_icon('2downarrow'),
            triggered=lambda: self.move_to(absolute=1))
        self.add_button = self.create_toolbutton(
            PathManagerToolbuttons.AddPath,
            tip=_('Add path'),
            icon=self.create_icon('edit_add'),
            triggered=lambda x: self.add_path())
        self.remove_button = self.create_toolbutton(
            PathManagerToolbuttons.RemovePath,
            tip=_('Remove path'),
            icon=self.create_icon('editclear'),
            triggered=lambda x: self.remove_path())
        self.export_button = self.create_toolbutton(
            PathManagerToolbuttons.ExportPaths,
            icon=self.create_icon('fileexport'),
            triggered=self.export_pythonpath,
            tip=_("Export to PYTHONPATH environment variable"))

        self.selection_widgets = [self.movetop_button, self.moveup_button,
                                  self.movedown_button, self.movebottom_button]
        return (
            [self.add_button, self.remove_button] +
            self.selection_widgets + [self.export_button]
        )

    def _create_item(self, path):
        """Helper to create a new list item."""
        item = QListWidgetItem(path)

        if path in self.project_path:
            item.setFlags(Qt.NoItemFlags | Qt.ItemIsUserCheckable)
            item.setCheckState(Qt.Checked)
        elif path in self.not_active_path:
            item.setFlags(item.flags() | Qt.ItemIsUserCheckable)
            item.setCheckState(Qt.Unchecked)
        else:
            item.setFlags(item.flags() | Qt.ItemIsUserCheckable)
            item.setCheckState(Qt.Checked)

        return item

    def _create_header(self, text):
        """Create a header for a given path section."""
        header_item = QListWidgetItem()
        header_widget = QLabel(text)

        # Disable item so we can remove its background color
        header_item.setFlags(header_item.flags() & ~Qt.ItemIsEnabled)

        # Header is centered
        header_widget.setAlignment(Qt.AlignHCenter)

        # Make header appear in bold
        font = header_widget.font()
        font.setBold(True)
        header_widget.setFont(font)

        # Increase height to make header stand over paths
        fm = QFontMetrics(font)
        header_item.setSizeHint(
            QSize(20, fm.capHeight() + 6 * AppStyle.MarginSize)
        )

        return header_item, header_widget

    @property
    def _stylesheet(self):
        """Style for the list of paths"""
        # This is necessary to match the buttons style with the rest of Spyder
        toolbar_stylesheet = PANES_TOOLBAR_STYLESHEET.get_copy()
        css = toolbar_stylesheet.get_stylesheet()

        css.QListView.setValues(
            padding=f"{AppStyle.MarginSize + 1}px"
        )

        css["QListView::item"].setValues(
            padding=f"{AppStyle.MarginSize + (1 if WIN else 0)}px"
        )

        css["QListView::item:disabled"].setValues(
            backgroundColor="transparent"
        )

        return css.toString()

    # ---- Public methods
    # -------------------------------------------------------------------------
    @property
    def editable_bottom_row(self):
        """Maximum bottom row count that is editable."""
        bottom_row = 0

        if self.project_header:
            bottom_row += len(self.project_path) + 1
        if self.user_header:
            bottom_row += len(self.user_path)

        return bottom_row

    @property
    def editable_top_row(self):
        """Maximum top row count that is editable."""
        top_row = 0

        if self.project_header:
            top_row += len(self.project_path) + 1
        if self.user_header:
            top_row += 1

        return top_row

    def setup(self):
        """Populate list widget."""
        self.listwidget.clear()
        self.headers.clear()
        self.project_header = None
        self.user_header = None
        self.system_header = None

        # Project path
        if self.project_path:
            self.project_header, project_widget = (
                self._create_header(_("Project path"))
            )
            self.headers.append(self.project_header)
            self.listwidget.addItem(self.project_header)
            self.listwidget.setItemWidget(self.project_header, project_widget)

            for path in self.project_path:
                item = self._create_item(path)
                self.listwidget.addItem(item)

        # Paths added by the user
        if self.user_path:
            self.user_header, user_widget = (
                self._create_header(_("User paths"))
            )
            self.headers.append(self.user_header)
            self.listwidget.addItem(self.user_header)
            self.listwidget.setItemWidget(self.user_header, user_widget)

            for path in self.user_path:
                item = self._create_item(path)
                self.listwidget.addItem(item)

        # System path
        if self.system_path:
            self.system_header, system_widget = (
                self._create_header(_("System PYTHONPATH"))
            )
            self.headers.append(self.system_header)
            self.listwidget.addItem(self.system_header)
            self.listwidget.setItemWidget(self.system_header, system_widget)

            for path in self.system_path:
                item = self._create_item(path)
                self.listwidget.addItem(item)

        self.listwidget.setCurrentRow(0)
        self.original_path_dict = self.get_path_dict()
        self.refresh()

    @Slot()
    def export_pythonpath(self):
        """
        Export to PYTHONPATH environment variable
        Only apply to: current user.
        """
        answer = QMessageBox.question(
            self,
            _("Export"),
            _("This will export Spyder's path list to the "
              "<b>PYTHONPATH</b> environment variable for the current user, "
              "allowing you to run your Python modules outside Spyder "
              "without having to configure sys.path. "
              "<br><br>"
              "Do you want to clear the contents of PYTHONPATH before "
              "adding Spyder's path list?"),
            QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel
        )

        if answer == QMessageBox.Cancel:
            return

        env = get_user_env()

        # This doesn't include the project path because it's a transient
        # directory, i.e. only used in Spyder and during specific
        # circumstances.
        active_path = [k for k, v in self.get_path_dict().items() if v]

        if answer == QMessageBox.Yes:
            ppath = active_path
        else:
            ppath = env.get('PYTHONPATH', [])
            if not isinstance(ppath, list):
                ppath = [ppath]

            ppath = [p for p in ppath if p not in active_path]
            ppath = ppath + active_path

        os.environ['PYTHONPATH'] = os.pathsep.join(ppath)

        # Update widget so changes are reflected on it immediately
        self.update_paths(system_path=tuple(ppath))
        self.set_conf('system_path', tuple(ppath))
        self.setup()

        env['PYTHONPATH'] = list(ppath)
        set_user_env(env, parent=self)

    def get_path_dict(self, project_path=False):
        """
        Return an ordered dict with the path entries as keys and the active
        state as the value.

        If `project_path` is True, its entries are also included.
        """
        odict = OrderedDict()
        for row in range(self.listwidget.count()):
            item = self.listwidget.item(row)
            path = item.text()
            if item not in self.headers:
                if path in self.project_path and not project_path:
                    continue
                odict[path] = item.checkState() == Qt.Checked

        return odict

    def get_user_path(self):
        """Get current user path as displayed on listwidget."""
        user_path = []
        for row in range(self.listwidget.count()):
            item = self.listwidget.item(row)
            path = item.text()
            if item not in self.headers:
                if path not in (self.project_path + self.system_path):
                    user_path.append(path)

        return user_path

    def update_paths(self, path=None, not_active_path=None, system_path=None):
        """Update path attributes."""
        if path is not None:
            self.path = path
        if not_active_path is not None:
            self.not_active_path = not_active_path
        if system_path is not None:
            self.system_path = system_path

        previous_system_path = self.get_conf('system_path', ())
        self.user_path = [
            path for path in self.path
            if path not in (self.system_path + previous_system_path)
        ]

    def refresh(self):
        """Refresh toolbar widgets."""
        current_item = self.listwidget.currentItem()
        enabled = current_item is not None
        for widget in self.selection_widgets:
            widget.setEnabled(enabled)

        # Main variables
        row = self.listwidget.currentRow()
        disable_widgets = []

        # Move up/top disabled for less than top editable item.
        if row <= self.editable_top_row:
            disable_widgets.extend([self.movetop_button, self.moveup_button])

        # Move down/bottom disabled for bottom item
        if row == self.editable_bottom_row:
            disable_widgets.extend([self.movebottom_button,
                                    self.movedown_button])

        # Disable almost all buttons on headers or system PYTHONPATH
        if current_item in self.headers or row > self.editable_bottom_row:
            disable_widgets.extend(
                [self.movetop_button, self.moveup_button,
                 self.movebottom_button, self.movedown_button]
            )

        for widget in disable_widgets:
            widget.setEnabled(False)

        # Enable remove button only for user paths
        self.remove_button.setEnabled(
            not current_item in self.headers
            and (self.editable_top_row <= row <= self.editable_bottom_row)
        )

        self.export_button.setEnabled(self.listwidget.count() > 0)

        # Ok button only enabled if actual changes occur
        self.button_ok.setEnabled(
            self.original_path_dict != self.get_path_dict()
        )

    @Slot()
    def add_path(self, directory=None):
        """
        Add path to list widget.

        If `directory` is provided, the folder dialog is overridden.
        """
        if directory is None:
            self.redirect_stdio.emit(False)
            directory = getexistingdirectory(self, _("Select directory"),
                                             self.last_path)
            self.redirect_stdio.emit(True)
            if not directory:
                return

        directory = osp.abspath(directory)
        self.last_path = directory

        if directory in self.get_path_dict():
            item = self.listwidget.findItems(directory, Qt.MatchExactly)[0]
            item.setCheckState(Qt.Checked)
            answer = QMessageBox.question(
                self,
                _("Add path"),
                _("This directory is already included in the list."
                  "<br> "
                  "Do you want to move it to the top of it?"),
                QMessageBox.Yes | QMessageBox.No)

            if answer == QMessageBox.Yes:
                item = self.listwidget.takeItem(self.listwidget.row(item))
                self.listwidget.insertItem(1, item)
                self.listwidget.setCurrentRow(1)
        else:
            if check_path(directory):
                if not self.user_header:
                    self.user_header, user_widget = (
                        self._create_header(_("User paths"))
                    )
                    self.headers.append(self.user_header)

                    if self.editable_top_row > 0:
                        header_row = self.editable_top_row - 1
                    else:
                        header_row = 0
                    self.listwidget.insertItem(header_row, self.user_header)
                    self.listwidget.setItemWidget(
                        self.user_header, user_widget
                    )

                # Add new path
                item = self._create_item(directory)
                self.listwidget.insertItem(self.editable_top_row, item)
                self.listwidget.setCurrentRow(self.editable_top_row)

                self.user_path.insert(0, directory)
            else:
                answer = QMessageBox.warning(
                    self,
                    _("Add path"),
                    _("This directory cannot be added to the path!"
                      "<br><br>"
                      "If you want to set a different Python interpreter, "
                      "please go to <tt>Preferences > Main interpreter</tt>"
                      "."),
                    QMessageBox.Ok)

        # Widget moves to back and loses focus on macOS,
        # see spyder-ide/spyder#20808
        if sys.platform == 'darwin':
            self.activateWindow()
            self.raise_()
            self.setFocus()

        self.refresh()

    @Slot()
    def remove_path(self, force=False):
        """
        Remove path from list widget.

        If `force` is True, the message box is overridden.
        """
        if self.listwidget.currentItem():
            if not force:
                answer = QMessageBox.warning(
                    self,
                    _("Remove path"),
                    _("Do you really want to remove the selected path?"),
                    QMessageBox.Yes | QMessageBox.No)

            if force or answer == QMessageBox.Yes:
                # Remove current item from user_path
                item = self.listwidget.currentItem()
                self.user_path.remove(item.text())

                # Remove selected item from view
                self.listwidget.takeItem(self.listwidget.currentRow())

                # Remove user header if there are no more user paths
                if len(self.user_path) == 0:
                    self.listwidget.takeItem(
                        self.listwidget.row(self.user_header)
                    )
                    self.headers.remove(self.user_header)
                    self.user_header = None

                # Refresh widget
                self.refresh()

    def move_to(self, absolute=None, relative=None):
        """Move items of list widget."""
        index = self.listwidget.currentRow()
        if absolute is not None:
            if absolute:
                new_index = self.editable_bottom_row
            else:
                new_index = self.editable_top_row
        else:
            new_index = index + relative

        new_index = max(1, min(self.editable_bottom_row, new_index))
        item = self.listwidget.takeItem(index)
        self.listwidget.insertItem(new_index, item)
        self.listwidget.setCurrentRow(new_index)

        self.user_path = self.get_user_path()
        self.refresh()

    def current_row(self):
        """Returns the current row of the list."""
        return self.listwidget.currentRow()

    def set_current_row(self, row):
        """Set the current row of the list."""
        self.listwidget.setCurrentRow(row)

    def row_check_state(self, row):
        """Return the checked state for item in row."""
        item = self.listwidget.item(row)
        return item.checkState()

    def set_row_check_state(self, row, value):
        """Set the current checked state for item in row."""
        item = self.listwidget.item(row)
        item.setCheckState(value)

    def count(self):
        """Return the number of items."""
        return self.listwidget.count()

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def _update_system_path(self):
        """
        Request to update path values on main window if current and previous
        system paths are different.
        """
        if self.system_path != self.get_conf('system_path', default=()):
            self.sig_path_changed.emit(self.get_path_dict())
        self.set_conf('system_path', self.system_path)

    def accept(self):
        """Override Qt method."""
        path_dict = self.get_path_dict()
        if self.original_path_dict != path_dict:
            self.sig_path_changed.emit(path_dict)
        super().accept()

    def reject(self):
        self._update_system_path()
        super().reject()

    def closeEvent(self, event):
        self._update_system_path()
        super().closeEvent(event)


def test():
    """Run path manager test."""
    from spyder.utils.qthelpers import qapplication

    _ = qapplication()
    dlg = PathManager(
        None,
        path=tuple(sys.path[:1]),
        project_path=tuple(sys.path[-2:]),
    )

    def callback(path_dict):
        sys.stdout.write(str(path_dict))

    dlg.sig_path_changed.connect(callback)
    sys.exit(dlg.exec_())


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Dialog to handle remote connections."""

# Standard library imports
from __future__ import annotations
from collections.abc import Iterable
import re
from typing import TypedDict
import uuid

# Third party imports
import qstylizer.style
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import (
    QDialogButtonBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QStackedWidget,
    QStackedLayout,
    QVBoxLayout,
    QWidget,
)

# Local imports
from spyder.api.fonts import SpyderFontType, SpyderFontsMixin
from spyder.api.translations import _
from spyder.api.utils import get_class_values
from spyder.api.widgets.dialogs import SpyderDialogButtonBox
from spyder.plugins.remoteclient.api.protocol import (
    ConnectionInfo,
    ConnectionStatus,
    RemoteClientLog,
    SSHClientOptions,
)
from spyder.plugins.remoteclient.widgets import AuthenticationMethod
from spyder.plugins.remoteclient.widgets.connectionstatus import (
    ConnectionStatusWidget,
)
from spyder.utils.icon_manager import ima
from spyder.utils.palette import SpyderPalette
from spyder.utils.stylesheet import AppStyle, MAC, WIN
from spyder.widgets.config import SpyderConfigPage
from spyder.widgets.helperwidgets import TipWidget
from spyder.widgets.sidebardialog import SidebarDialog


# =============================================================================
# ---- Constants
# =============================================================================
class ValidationReasons(TypedDict):
    repeated_name: bool | None
    missing_info: bool | None
    invalid_address: bool | None


# =============================================================================
# ---- Auxiliary widgets
# =============================================================================
class ValidationLabel(QLabel):
    """Label to report to users that info failed to be validated."""

    def __init__(self, parent):
        super().__init__("", parent)

        # Set main attributes
        self.setWordWrap(True)
        self.setVisible(False)

        # Set style
        css = qstylizer.style.StyleSheet()
        css.QLabel.setValues(
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_2,
            # Top margin is set by the layout
            marginTop="0px",
            marginRight=f"{9 * AppStyle.MarginSize}px",
            # We don't need bottom margin because there are no other elements
            # below this one.
            marginBottom="0px",
            # The extra 5px are necessary because we need to add them to all
            # lineedits in this dialog to align them to the labels on top of
            # them (see SpyderConfigPage.create_lineedit).
            marginLeft=f"{9 * AppStyle.MarginSize + 5}px",
            padding=f"{3 * AppStyle.MarginSize}px {6 * AppStyle.MarginSize}px",
            borderRadius=SpyderPalette.SIZE_BORDER_RADIUS,
        )

        self.setStyleSheet(css.toString())

    def set_text(self, reasons: ValidationReasons):
        n_reasons = list(reasons.values()).count(True)
        prefix = "- " if n_reasons > 1 else ""
        suffix = "<br>" if n_reasons > 1 else ""

        text = ""
        if reasons.get("repeated_name"):
            text += (
                prefix
                + _(
                    "The name you selected is already used by another "
                    "connection."
                )
                + suffix
            )

        if reasons.get("invalid_address"):
            text += (
                prefix
                + _(
                    "The address you provided is not a valid IP or domain "
                    "name."
                )
                + suffix
            )

        if reasons.get("missing_info"):
            text += (
                prefix
                + _("There are missing fields on this page.")
            )

        self.setAlignment(Qt.AlignCenter if n_reasons == 1 else Qt.AlignLeft)
        self.setText(text)


# =============================================================================
# ---- Pages
# =============================================================================
class BaseConnectionPage(SpyderConfigPage, SpyderFontsMixin):
    """Base class to create connection pages."""

    MIN_HEIGHT = 450
    NEW_CONNECTION = False
    CONF_SECTION = "remoteclient"

    def __init__(self, parent, host_id=None):
        super().__init__(parent)

        # host_id is None only for the new connection page
        if host_id is None:
            self.host_id = str(uuid.uuid4())
            self.status = ConnectionStatus.Inactive
        else:
            self.host_id = host_id
            self.status = self.get_option(
                f"{host_id}/status", default=ConnectionStatus.Inactive
            )

        self._widgets_for_validation = {}
        self._validation_labels = {}
        self._name_widgets = {}
        self._address_widgets = {}

    # ---- Public API
    # -------------------------------------------------------------------------
    def auth_method(self, from_gui=False):
        if from_gui:
            if self._auth_methods.combobox.currentIndex() == 0:
                auth_method = AuthenticationMethod.Password
            elif self._auth_methods.combobox.currentIndex() == 1:
                auth_method = AuthenticationMethod.KeyFile
            else:
                auth_method = AuthenticationMethod.ConfigFile
        else:
            auth_method = self.get_option(f"{self.host_id}/auth_method")

        return auth_method

    def validate_page(self):
        """Validate contents before saving the connection."""
        # Hide label and clear status actions from all pages!

        # Get widgets we're going to interact with
        auth_method = self.auth_method(from_gui=True)
        widgets = self._widgets_for_validation[auth_method]
        validate_label = self._validation_labels[auth_method]

        # Hide label in case the validation pass
        validate_label.setVisible(False)

        reasons: ValidationReasons = {}
        for widget in widgets:
            if not widget.textbox.text():
                # Validate that the required fields are not empty
                widget.status_action.setVisible(True)
                widget.status_action.setToolTip(_("This field is empty"))
                reasons["missing_info"] = True
            elif widget == self._name_widgets[auth_method]:
                # Validate the server name is different from the ones already
                # introduced
                widget.status_action.setVisible(False)
                current_name = widget.textbox.text()
                if not self._validate_name(current_name):
                    reasons["repeated_name"] = True
                    widget.status_action.setVisible(True)
            elif widget == self._address_widgets.get(auth_method):
                # Validate address
                widget.status_action.setVisible(False)
                address = widget.textbox.text()
                if not self._validate_address(address):
                    reasons["invalid_address"] = True
                    widget.status_action.setVisible(True)
            else:
                widget.status_action.setVisible(False)

        if reasons:
            validate_label.set_text(reasons)
            validate_label.setVisible(True)

        return False if reasons else True

    def create_connection_info_widget(self):
        """
        Create widget that contains all other widgets to receive or display
        connection info.
        """
        # Show intro text and tip for new connections
        if self.NEW_CONNECTION:
            # Widgets
            intro_label = QLabel(
                _("Configure SSH settings for connecting to remote hosts")
            )
            intro_tip_text = _(
                "Spyder will use this connection to start remote kernels in "
                "the IPython Console. This allows you to use Spyder locally "
                "while running code remotely, such as on a cloud instance, "
                "office workstation or high-performance cluster."
            )
            intro_tip = TipWidget(
                tip_text=intro_tip_text,
                icon=ima.icon('info_tip'),
                hover_icon=ima.icon('info_tip_hover'),
                size=AppStyle.ConfigPageIconSize + 2,
                wrap_text=True,
            )

            # Increase font size to make it more relevant
            font = self.get_font(SpyderFontType.Interface)
            font.setPointSize(font.pointSize() + 1)
            intro_label.setFont(font)

            # Layout
            intro_layout = QHBoxLayout()
            intro_layout.setContentsMargins(0, 0, 0, 0)
            intro_layout.setSpacing(0)
            intro_layout.setAlignment(Qt.AlignCenter)
            intro_layout.addWidget(intro_label)
            intro_layout.addWidget(intro_tip)

        # Authentication methods
        # TODO: The config file method is not implemented yet, so we need to
        # disable it for now.
        methods = (
            (_('Password'), AuthenticationMethod.Password),
            (_('Key file'), AuthenticationMethod.KeyFile),
            # (_('Configuration file'), AuthenticationMethod.ConfigFile),
        )

        self._auth_methods = self.create_combobox(
            _("Authentication method:"),
            methods,
            f"{self.host_id}/auth_method"
        )

        # Subpages
        password_subpage = self._create_password_subpage()
        keyfile_subpage = self._create_keyfile_subpage()
        configfile_subpage = self._create_configfile_subpage()

        subpages = QStackedWidget(self)
        subpages.addWidget(password_subpage)
        subpages.addWidget(keyfile_subpage)
        subpages.addWidget(configfile_subpage)

        # Signals
        self._auth_methods.combobox.currentIndexChanged.connect(
            subpages.setCurrentIndex
        )

        # Show password subpage by default for new connections
        if self.NEW_CONNECTION:
            self._auth_methods.combobox.setCurrentIndex(0)

        # Final layout
        layout = QVBoxLayout()
        layout.setContentsMargins(
            3 * AppStyle.MarginSize, 0, 3 * AppStyle.MarginSize, 0
        )
        if self.NEW_CONNECTION:
            layout.addLayout(intro_layout)
            layout.addSpacing(8 * AppStyle.MarginSize)
        layout.addWidget(self._auth_methods)
        layout.addSpacing(5 * AppStyle.MarginSize)
        layout.addWidget(subpages)

        connection_info_widget = QWidget(self)
        connection_info_widget.setLayout(layout)

        return connection_info_widget

    # ---- Private API
    # -------------------------------------------------------------------------
    def _create_common_elements(self, auth_method):
        """Common elements for the password and keyfile subpages."""
        # Widgets
        name = self.create_lineedit(
            text=_("Name *"),
            option=f"{self.host_id}/{auth_method}/name",
            tip=_("Introduce a name to identify your connection"),
            validate_callback=self._validate_name,
            validate_reason=_("This connection name is already taken"),
        )

        address = self.create_lineedit(
            text=_("Remote address *"),
            option=f"{self.host_id}/{auth_method}/address",
            tip=_(
                "This is the IP address or domain name of your remote machine"
            ),
            validate_callback=self._validate_address,
            validate_reason=_("The address is not a valid IP or domain name"),
        )

        port = self.create_spinbox(
            prefix=_("Port"),
            suffix="",
            option=f"{self.host_id}/{auth_method}/port",
            min_=1,
            max_=65535
        )
        port.spinbox.setStyleSheet("margin-left: 5px")

        username = self.create_lineedit(
            text=_("Username *"),
            option=f"{self.host_id}/{auth_method}/username",
            status_icon=ima.icon("error"),
        )

        self._widgets_for_validation[f"{auth_method}"] = [
            name,
            address,
            username,
        ]
        self._name_widgets[f"{auth_method}"] = name
        self._address_widgets[f"{auth_method}"] = address

        # Set 22 as the default port for new conenctions
        if not self.LOAD_FROM_CONFIG:
            port.spinbox.setValue(22)

        # Hide the container widgets because we only use their components
        address.hide()
        port.hide()

        # Layout for the address label
        address_label_layout = QHBoxLayout()
        address_label_layout.setSpacing(0)
        address_label_layout.addWidget(address.label)
        address_label_layout.addWidget(address.help_label)
        address_label_layout.addStretch()

        # Address layout
        address_layout = QGridLayout()
        address_layout.setContentsMargins(0, 0, 0, 0)

        address_layout.addLayout(address_label_layout, 0, 0)
        address_layout.addWidget(address.textbox, 1, 0)
        address_layout.addWidget(port.plabel, 0, 1)
        address_layout.addWidget(port.spinbox, 1, 1)

        return name, address_layout, username

    def _create_password_subpage(self):
        # Widgets
        name, address_layout, username = self._create_common_elements(
            auth_method=AuthenticationMethod.Password
        )

        password = self.create_lineedit(
            text=_("Password *"),
            option=f"{self.host_id}/password",
            tip=(
                _("Your password will be saved securely by Spyder")
                if self.NEW_CONNECTION
                else _("Your password is saved securely by Spyder")
            ),
            status_icon=ima.icon("error"),
            password=True
        )

        validation_label = ValidationLabel(self)

        # Add widgets to their required dicts
        self._widgets_for_validation[AuthenticationMethod.Password].append(
            password
        )

        self._validation_labels[
            AuthenticationMethod.Password
        ] = validation_label

        # Layout
        password_layout = QVBoxLayout()
        password_layout.setContentsMargins(0, 0, 0, 0)

        password_layout.addWidget(name)
        password_layout.addSpacing(5 * AppStyle.MarginSize)
        password_layout.addLayout(address_layout)
        password_layout.addSpacing(5 * AppStyle.MarginSize)
        password_layout.addWidget(username)
        password_layout.addSpacing(5 * AppStyle.MarginSize)
        password_layout.addWidget(password)
        password_layout.addSpacing(7 * AppStyle.MarginSize)
        password_layout.addWidget(validation_label)
        password_layout.addStretch()

        password_subpage = QWidget(self)
        password_subpage.setLayout(password_layout)

        return password_subpage

    def _create_keyfile_subpage(self):
        # Widgets
        name, address_layout, username = self._create_common_elements(
            auth_method=AuthenticationMethod.KeyFile
        )

        keyfile = self.create_browsefile(
            text=_("Key file *"),
            option=f"{self.host_id}/keyfile",
            alignment=Qt.Vertical,
            status_icon=ima.icon("error"),
        )

        passpharse = self.create_lineedit(
            text=_("Passpharse"),
            option=f"{self.host_id}/passpharse",
            tip=(
                _("Your passphrase will be saved securely by Spyder")
                if self.NEW_CONNECTION
                else _("Your passphrase is saved securely by Spyder")
            ),
            password=True
        )

        validation_label = ValidationLabel(self)

        # Add widgets to their required dicts
        self._widgets_for_validation[AuthenticationMethod.KeyFile].append(
            keyfile
        )

        self._validation_labels[
            AuthenticationMethod.KeyFile
        ] = validation_label

        # Layout
        keyfile_layout = QVBoxLayout()
        keyfile_layout.setContentsMargins(0, 0, 0, 0)

        keyfile_layout.addWidget(name)
        keyfile_layout.addSpacing(5 * AppStyle.MarginSize)
        keyfile_layout.addLayout(address_layout)
        keyfile_layout.addSpacing(5 * AppStyle.MarginSize)
        keyfile_layout.addWidget(username)
        keyfile_layout.addSpacing(5 * AppStyle.MarginSize)
        keyfile_layout.addWidget(keyfile)
        keyfile_layout.addSpacing(5 * AppStyle.MarginSize)
        keyfile_layout.addWidget(passpharse)
        keyfile_layout.addSpacing(7 * AppStyle.MarginSize)
        keyfile_layout.addWidget(validation_label)
        keyfile_layout.addStretch()

        keyfile_subpage = QWidget(self)
        keyfile_subpage.setLayout(keyfile_layout)

        return keyfile_subpage

    def _create_configfile_subpage(self):
        # Widgets
        name = self.create_lineedit(
            text=_("Name *"),
            option=f"{self.host_id}/{AuthenticationMethod.ConfigFile}/name",
            tip=_("Introduce a name to identify your connection"),
            status_icon=ima.icon("error"),
        )

        configfile = self.create_browsefile(
            text=_("Configuration file *"),
            option=f"{self.host_id}/configfile",
            alignment=Qt.Vertical,
            status_icon=ima.icon("error"),
        )

        validation_label = ValidationLabel(self)

        # Add widgets to their required dicts
        self._name_widgets[AuthenticationMethod.ConfigFile] = name
        self._widgets_for_validation[AuthenticationMethod.ConfigFile] = [
            name,
            configfile,
        ]
        self._validation_labels[
            AuthenticationMethod.ConfigFile
        ] = validation_label

        # Layout
        configfile_layout = QVBoxLayout()
        configfile_layout.setContentsMargins(0, 0, 0, 0)
        configfile_layout.addWidget(name)
        configfile_layout.addSpacing(5 * AppStyle.MarginSize)
        configfile_layout.addWidget(configfile)
        configfile_layout.addSpacing(7 * AppStyle.MarginSize)
        configfile_layout.addWidget(validation_label)
        configfile_layout.addStretch()

        configfile_widget = QWidget(self)
        configfile_widget.setLayout(configfile_layout)

        return configfile_widget

    def _validate_name(self, name):
        """Check connection name is not taken by a previous connection."""
        servers = self.get_option("servers", default={})
        for server in servers:
            # Don't check repeated name with the same server
            if server == self.host_id:
                continue

            names = [
                self.get_option(f"{server}/{method}/name")
                for method in get_class_values(AuthenticationMethod)
            ]

            if name in names:
                return False

        return True

    def _validate_address(self, address):
        """Validate if address introduced by users is correct."""
        # Regex pattern for a valid domain name (simplified version)
        domain_pattern = (
            r'^([a-zA-Z0-9][a-zA-Z0-9-]{0,61}[a-zA-Z0-9]\.){1,}[a-zA-Z]{2,}$'
        )

        # Regex pattern for a valid IPv4 address
        ipv4_pattern = r'^(\d{1,3}\.){3}\d{1,3}$'

        # Regex pattern for a valid IPv6 address (simplified version)
        ipv6_pattern = r'^([\da-fA-F]{1,4}:){7}[\da-fA-F]{1,4}$'

        # Combined pattern to check all three formats
        combined_pattern = (
            f'({domain_pattern})|({ipv4_pattern})|({ipv6_pattern})'
        )

        address_re = re.compile(combined_pattern)
        return True if address_re.match(address) else False


class NewConnectionPage(BaseConnectionPage):
    """Page to receive SSH credentials for a remote connection."""

    MAX_WIDTH = 540 if MAC else 520
    LOAD_FROM_CONFIG = False
    NEW_CONNECTION = True

    # ---- SidebarPage API
    # -------------------------------------------------------------------------
    def get_name(self):
        return _("New connection")

    def setup_page(self):
        info_widget = self.create_connection_info_widget()

        # Use a stacked layout so we can hide the current widgets and create
        # new ones in case users want to introduce more connections.
        self.layout = QStackedLayout()
        self.layout.addWidget(info_widget)
        self.setLayout(self.layout)

    def get_icon(self):
        return self.create_icon("add_server")

    # ---- Public API
    # -------------------------------------------------------------------------
    def reset_page(self, clear=False):
        """Reset page to allow users to introduce a new connection."""
        # Set a new host id
        self.host_id = str(uuid.uuid4())

        if clear:
            # Reset tracked widgets
            self.reset_widget_dicts()

            # Add a new, clean set of widgets to the page
            clean_info_widget = self.create_connection_info_widget()
            self.layout.addWidget(clean_info_widget)
            self.layout.setCurrentWidget(clean_info_widget)
        else:
            # Change option names associated to all widgets present in the page
            # to reference the new host_id
            for widgets in [self.comboboxes, self.lineedits, self.spinboxes]:
                for widget in widgets:
                    section, option, default = widgets[widget]
                    new_option = "/".join(
                        [self.host_id] + option.split("/")[1:]
                    )
                    widgets[widget] = (section, new_option, default)


class ConnectionPage(BaseConnectionPage):
    """Page to display connection status and info for a remote machine."""

    MAX_WIDTH = 600 if MAC else 580

    def __init__(self, parent, host_id):
        super().__init__(parent, host_id)
        self.new_name = None

    # ---- SidebarPage API
    # -------------------------------------------------------------------------
    def get_name(self):
        return self.get_option(f"{self.host_id}/{self.auth_method()}/name")

    def setup_page(self):
        info_widget = self.create_connection_info_widget()
        self.status_widget = ConnectionStatusWidget(self, self.host_id)

        self.create_tab(_("Connection status"), self.status_widget)
        self.create_tab(_("Connection info"), info_widget)

    def get_icon(self):
        return self.create_icon("remote_server")

    # ---- Public API
    # -------------------------------------------------------------------------
    def save_server_info(self):
        # Mapping from options in our config system to those accepted by
        # asyncssh
        options = SSHClientOptions(
            host=self.get_option(
                f"{self.host_id}/{self.auth_method()}/address"
            ),
            port=self.get_option(f"{self.host_id}/{self.auth_method()}/port"),
            username=self.get_option(
                f"{self.host_id}/{self.auth_method()}/username"
            ),
            client_keys=self.get_option(f"{self.host_id}/keyfile"),
            config=self.get_option(f"{self.host_id}/configfile"),
        )

        servers = self.get_option("servers", default={})
        servers[self.host_id] = options
        self.set_option("servers", servers)

    def remove_config_options(self):
        """Remove config options associated to this connection."""
        # Remove current server from the dict of them
        servers = self.get_option("servers")
        servers.pop(self.host_id)
        self.set_option("servers", servers)

        # Remove regular options
        options = [
            "auth_method",
            "password_login/name",
            "password_login/address",
            "password_login/port",
            "password_login/username",
            "keyfile_login/name",
            "keyfile_login/address",
            "keyfile_login/port",
            "keyfile_login/username",
            "keyfile",
            "configfile",
        ]
        for option in options:
            self.remove_option(f"{self.host_id}/{option}")

        # Remove secure options
        for secure_option in ["password", "passpharse"]:
            # One of these options was saved securely and other as empty in our
            # config system, so we try to remove them both.
            for secure in [True, False]:
                self.remove_option(
                    f"{self.host_id}/{secure_option}", secure=secure
                )

    def update_status(self, info: ConnectionInfo):
        if info["id"] == self.host_id:
            self.status = info["status"]
            self.status_widget.update_status(info)

    def add_log(self, log: RemoteClientLog):
        if log["id"] == self.host_id:
            self.status_widget.add_log(log)

    def add_logs(self, logs: Iterable):
        self.status_widget.add_logs(logs)

    def has_new_name(self):
        """Check if users changed the connection name."""
        current_auth_method = self.auth_method(from_gui=True)
        current_name = self._name_widgets[current_auth_method].textbox.text()

        if self.get_name() != current_name:
            self.new_name = current_name
            return True
        else:
            self.new_name = None
            return False


# =============================================================================
# ---- Dialog
# =============================================================================
class ConnectionDialog(SidebarDialog):
    """
    Dialog to handle and display remote connection information for different
    machines.
    """

    TITLE = _("Remote connections")
    MIN_WIDTH = 900 if MAC else (830 if WIN else 880)
    MIN_HEIGHT = 700 if MAC else (635 if WIN else 650)
    PAGE_CLASSES = [NewConnectionPage]

    sig_start_server_requested = Signal(str)
    sig_stop_server_requested = Signal(str)
    sig_server_renamed = Signal(str)
    sig_connections_changed = Signal()

    def __init__(self, parent=None):
        self.ICON = ima.icon('remote_server')
        super().__init__(parent)
        self._container = parent

        # -- Setup
        self._add_saved_connection_pages()

        # If there's more than one page, give focus to the first server because
        # users will probably want to interact with servers here rather than
        # create new connections.
        if self.number_of_pages() > 1:
            # Index 1 is the separator added after the new connection page
            self.set_current_index(2)

        # -- Signals
        if self._container is not None:
            self._container.sig_connection_status_changed.connect(
                self._update_connection_buttons_state
            )

    # ---- SidebarDialog API
    # -------------------------------------------------------------------------
    def create_buttons(self):
        bbox = SpyderDialogButtonBox(QDialogButtonBox.Cancel)

        self._button_save_connection = QPushButton(_("Save connection"))
        self._button_save_connection.clicked.connect(
            self._save_connection_info
        )
        bbox.addButton(
            self._button_save_connection, QDialogButtonBox.ResetRole
        )

        self._button_remove_connection = QPushButton(_("Remove connection"))
        self._button_remove_connection.clicked.connect(
            self._remove_connection_info
        )
        bbox.addButton(
            self._button_remove_connection, QDialogButtonBox.ResetRole
        )

        self._button_clear_settings = QPushButton(_("Clear settings"))
        self._button_clear_settings.clicked.connect(self._clear_settings)
        bbox.addButton(
            self._button_clear_settings, QDialogButtonBox.ActionRole
        )

        self._button_connect = QPushButton(_("Connect"))
        self._button_connect.clicked.connect(self._start_server)
        bbox.addButton(self._button_connect, QDialogButtonBox.ActionRole)

        self._button_stop = QPushButton(_("Stop"))
        self._button_stop.clicked.connect(self._stop_server)
        bbox.addButton(self._button_stop, QDialogButtonBox.ActionRole)

        layout = QHBoxLayout()
        layout.addWidget(bbox)

        return bbox, layout

    def current_page_changed(self, index):
        """Update the state of buttons when moving from page to page."""
        page = self.get_page(index)
        if page.NEW_CONNECTION:
            self._button_save_connection.setEnabled(True)
            self._button_clear_settings.setHidden(False)
            self._button_remove_connection.setHidden(True)
            self._button_stop.setHidden(True)
        else:
            if page.is_modified:
                self._button_save_connection.setEnabled(True)
            else:
                self._button_save_connection.setEnabled(False)

            self._button_clear_settings.setHidden(True)
            self._button_remove_connection.setHidden(False)
            self._button_stop.setHidden(False)

        if page.status in [
            ConnectionStatus.Inactive,
            ConnectionStatus.Error,
        ]:
            self._button_connect.setEnabled(True)
        else:
            self._button_connect.setEnabled(False)

        # TODO: Check if it's possible to stop a connection while it's
        # connecting
        if page.status == ConnectionStatus.Active:
            self._button_stop.setEnabled(True)
        else:
            self._button_stop.setEnabled(False)

    # ---- Private API
    # -------------------------------------------------------------------------
    def _save_connection_info(self):
        """Save the connection info stored in a page."""
        page = self.get_page()

        # Validate info
        if not page.validate_page():
            return

        if page.NEW_CONNECTION:
            # Save info provided by users
            page.save_to_conf()

            # Add separator if needed
            if self.number_of_pages() == 1:
                self.add_separator()

            # Add connection page to the dialog with the new info
            self._add_connection_page(host_id=page.host_id, new=True)

            # Give focus to the new page
            self.set_current_index(self.number_of_pages() - 1)

            # Reset page in case users want to introduce another connection
            page.reset_page()

            # Inform container that a change in connections took place
            self.sig_connections_changed.emit()
        else:
            # Update name in the dialog if it was changed by users. This needs
            # to be done before calling save_to_conf so that we can compare the
            # saved name with the current one.
            if page.has_new_name():
                self.get_item().setText(page.new_name)

            # Update connection info
            page.save_to_conf()

            # After saving to our config system, we can inform the container
            # that a change in connections took place.
            if page.new_name is not None:
                self.sig_connections_changed.emit()
                self.sig_server_renamed.emit(page.host_id)
                page.new_name = None

            # Mark page as not modified and disable save button
            page.set_modified(False)
            self._button_save_connection.setEnabled(False)

    def _remove_connection_info(self):
        """
        Remove the connection info stored in a given page and hide it as well.
        """
        page = self.get_page()
        if not page.NEW_CONNECTION:
            reply = QMessageBox.question(
                self,
                _("Remove connection"),
                _(
                    "Do you want to remove the connection called <b>{}</b>?"
                ).format(page.get_name()),
                QMessageBox.Yes,
                QMessageBox.No,
            )

            if reply == QMessageBox.Yes:
                self.hide_page()
                page.remove_config_options()

                # Inform container that a change in connections took place
                self.sig_connections_changed.emit()

    def _clear_settings(self):
        """Clear the setting introduced in the new connection page."""
        page = self.get_page()
        if page.NEW_CONNECTION:
            page.reset_page(clear=True)

    def _start_server(self):
        """Start the server corresponding to a given page."""
        page = self.get_page()

        # Validate info
        if not page.validate_page():
            return

        # This uses the current host_id in case users want to start a
        # connection directly from the new connection page (
        # _save_connection_info generates a new id fo that page at the end).
        host_id = page.host_id

        if page.NEW_CONNECTION or page.is_modified:
            # Save connection info if necessary
            self._save_connection_info()

            # TODO: Handle the case when the connection info is active and
            # users change its info.

        self.sig_start_server_requested.emit(host_id)

    def _stop_server(self):
        """Stop the server corresponding to a given page."""
        page = self.get_page()

        # The stop button is not visible in the new connection page
        if not page.NEW_CONNECTION:
            self._button_stop.setEnabled(False)
            self.sig_stop_server_requested.emit(page.host_id)

    def _add_connection_page(self, host_id: str, new: bool):
        """Add a new connection page to the dialog."""
        page = ConnectionPage(self, host_id=host_id)

        # This is necessary to make button_save_connection enabled when there
        # are config changes in the page
        page.apply_button_enabled.connect(
            self._update_button_save_connection_state
        )

        if new:
            page.save_server_info()

        self.add_page(page)

        # Add saved logs to the page
        if self._container is not None:
            page.add_logs(self._container.client_logs.get(host_id, []))

            # This updates the info shown in the "Connection info" tab of pages
            self._container.sig_connection_status_changed.connect(
                page.update_status
            )
            self._container.sig_client_message_logged.connect(page.add_log)

    def _add_saved_connection_pages(self):
        """Add a connection page for each server saved in our config system."""
        page = self.get_page(index=0)
        servers = page.get_option("servers", default={})

        if servers:
            self.add_separator()

            for id_ in servers.keys():
                self._add_connection_page(host_id=id_, new=False)

    def _update_button_save_connection_state(self, state: bool):
        """Update the state of the 'Save connection' button."""
        self._button_save_connection.setEnabled(state)

    def _update_connection_buttons_state(self, info: ConnectionInfo):
        """Update the state of the 'Connect' button."""
        page = self.get_page()
        if page.host_id == info["id"]:
            if info["status"] in [
                ConnectionStatus.Inactive,
                ConnectionStatus.Error,
            ]:
                self._button_connect.setEnabled(True)
            else:
                self._button_connect.setEnabled(False)

            # TODO: Check if it's possible to stop a connection while it's
            # connecting
            if info["status"] == ConnectionStatus.Active:
                self._button_stop.setEnabled(True)
            else:
                self._button_stop.setEnabled(False)


def test():
    from spyder.utils.qthelpers import qapplication
    from spyder.utils.stylesheet import APP_STYLESHEET

    app = qapplication()  # noqa
    app.setStyleSheet(str(APP_STYLESHEET))

    dialog = ConnectionDialog()
    dialog.exec_()


if __name__ == "__main__":
    test()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------


# Standard library imports
import os.path as osp
import time

# Third party imports
import rope.base.project
import rope.base.libutils
import rope.contrib.codeassist


ROPE_PREFS = {
               'ignore_syntax_errors': True,
               'ignore_bad_imports': True,
               'automatic_soa': False,
               'perform_doa': False,
               'import_dynload_stdmods': False,
               'soa_followed_calls': 2,
               'extension_modules': [
         "PyQt4", "PyQt4.QtGui", "QtGui", "PyQt4.QtCore", "QtCore",
         "PyQt4.QtScript", "QtScript", "os.path", "numpy", "scipy", "PIL",
         "OpenGL", "array", "audioop", "binascii", "cPickle", "cStringIO",
         "cmath", "collections", "datetime", "errno", "exceptions", "gc",
         "imageop", "imp", "itertools", "marshal", "math", "mmap", "msvcrt",
         "nt", "operator", "os", "parser", "rgbimg", "signal", "strop", "sys",
         "thread", "time", "wx", "wxPython", "xxsubtype", "zipimport", "zlib"],
               }

def ropetest():
    project = rope.base.project.Project('src', **ROPE_PREFS)
    project.validate(project.root)

    filename = osp.join('src', 'script.py')
    source_code = file(filename, 'rb').read()
    offset = len(source_code)

    resource = rope.base.libutils.path_to_resource(project, filename)

    t0 = time.time()

    proposals = rope.contrib.codeassist.code_assist(project, source_code,
                                                    offset, resource)
    proposals = rope.contrib.codeassist.sorted_proposals(proposals)

    print "%s: %d ms" % ("completion", 10*round(1e2*(time.time()-t0)))
    print 'loadtxt' in [proposal.name for proposal in proposals]

pydocextractor = rope.contrib.codeassist.PyDocExtractor()

def get_pyobject(project, source_code, offset, resource=None, maxfixes=1):
    fixer = rope.contrib.codeassist.fixsyntax.FixSyntax(project.pycore,
                                        source_code, resource, maxfixes)
    pyname = fixer.pyname_at(offset)
    if pyname is None:
        return None
    return pyname.get_object()

def get_calltip_from_pyobject(pyobject,
                              ignore_unknown=False, remove_self=False):
    return pydocextractor.get_calltip(pyobject, ignore_unknown, remove_self)

def get_doc_from_pyobject(pyobject):
    return pydocextractor.get_doc(pyobject)


from spyder import rope_patch
rope_patch.apply()

def other_features():
    project = rope.base.project.Project('src', **ROPE_PREFS)
    project.validate(project.root)

    filename = osp.join('src', 'script2.py')
    source_code = file(filename, 'rb').read()
    offset = len(source_code)

    resource = rope.base.libutils.path_to_resource(project, filename)

    t0 = time.time()

    cts = rope.contrib.codeassist.get_calltip(
                                    project, source_code, offset, resource)
    doc_text = rope.contrib.codeassist.get_doc(
                                    project, source_code, offset, resource)
    def_loc = rope.contrib.codeassist.get_definition_location(
                                    project, source_code, offset, resource)

    msg = "Testing other rope instrospection features"
    print msg
    print "="*len(msg)
    print ""
    print "%s: %d ms" % ("elapsed time", 10*round(1e2*(time.time()-t0)))
    print ""
    print 'calltip:', cts
    print 'definition location:', def_loc
    print 'doc:'
    print '*** DOCUMENTATION ***' + '*'*60
    print doc_text
    print '*********************' + '*'*60


if __name__ == '__main__':
    # ropetest()
    other_features()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License


# Standard library imports
import os
import sys

# Third party imports
import pytest
from qtpy.QtCore import QPoint, Qt
from qtpy.QtGui import QFont

# Local imports
from spyder.plugins.editor.widgets.codeeditor import CodeEditor
from spyder.plugins.debugger.utils.breakpointsmanager import BreakpointsManager

# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------

@pytest.fixture
def editor_bot(qtbot):
    widget = CodeEditor(None)
    widget.setup_editor(linenumbers=True,
                        markers=True,
                        show_blanks=True,
                        scrollflagarea=True,
                        font=QFont("Courier New", 10),
                        color_scheme='Zenburn',
                        language='Python')
    qtbot.addWidget(widget)
    return widget


# ---------------------------------------------------------------------------
# Examples to Test Against
# ---------------------------------------------------------------------------
long_code = """Line1
Line2
Line3
Line4
Line5
Line6
Line7
Line8
Line9
Line10
Line11
Line12
Line13
Line14
Line15
Line16
Line17
Line18
Line19
Line20
"""

short_code = """line1: Occurences
line2: Breakpoints
line3: TODOs
line4: Code Analysis: warning
line5: Code Analysis: error
line6: Found Results
"""


# ---------------------------------------------------------------------------
# Tests
# ---------------------------------------------------------------------------
def test_enabled(editor_bot):
    """"Test that enabling and disabling the srollflagarea panel make
    it visible or invisible depending on the case."""
    editor = editor_bot
    sfa = editor.scrollflagarea
    editor.show()
    editor.set_text(short_code)

    assert sfa.isVisible()
    sfa.set_enabled(False)
    assert not sfa.isVisible()


@pytest.mark.skipif(sys.platform.startswith('linux'),
                    reason="Fails in Linux")
def test_flag_painting(editor_bot, qtbot):
    """"Test that there is no error when painting all flag types on the
    scrollbar area when the editor vertical scrollbar is visible and not
    visible. There is seven different flags: breakpoints, todos, warnings,
    errors, found_results, and occurences"""
    editor = editor_bot
    editor.filename = "file.py"
    editor.breakpoints_manager = BreakpointsManager(editor)
    sfa = editor.scrollflagarea

    editor.resize(450, 300)
    editor.show()

    # Set a short text in the editor and assert that the slider is not visible.
    editor.set_text(short_code)
    qtbot.waitUntil(lambda: not sfa.slider)

    # Trigger the painting of all flag types.
    editor.breakpoints_manager.toogle_breakpoint(line_number=2)
    editor.process_todo([[True, 3]])
    analysis = [{'source': 'pycodestyle', 'range':{
                    'start': {'line': 4, 'character': 0},
                    'end': {'line': 4, 'character': 1}},
                 'line': 4, 'code': 'E227','message': 'E227 warning',
                 'severity': 2},
                 {'source': 'pyflakes', 'range':{
                     'start': {'line': 5, 'character': 0},
                     'end': {'line': 5, 'character': 1}},
                  'message': 'syntax error', 'severity': 1}]
    editor.process_code_analysis(analysis)
    editor.highlight_found_results('line6')
    with qtbot.waitSignal(editor.sig_flags_changed, raising=True,
                          timeout=5000):
        cursor = editor.textCursor()
        cursor.setPosition(2)
        editor.setTextCursor(cursor)

    # Set a long text in the editor and assert that the slider is visible.
    editor.set_text(long_code)

    # Trigger the painting of all flag types.
    editor.breakpoints_manager.toogle_breakpoint(line_number=2)
    editor.process_todo([[True, 3]])
    analysis = [{'source': 'pycodestyle', 'range':{
                    'start': {'line': 4, 'character': 0},
                    'end': {'line': 4, 'character': 1}},
                 'line': 4, 'code': 'E227','message': 'E227 warning',
                 'severity': 2},
                 {'source': 'pyflakes', 'range':{
                     'start': {'line': 5, 'character': 0},
                     'end': {'line': 5, 'character': 1}},
                  'message': 'syntax error', 'severity': 1}]
    editor.process_code_analysis(analysis)
    editor.highlight_found_results('line6')
    with qtbot.waitSignal(editor.sig_flags_changed, raising=True,
                          timeout=5000):
        cursor = editor.textCursor()
        cursor.setPosition(2)
        editor.setTextCursor(cursor)


def test_range_indicator_visible_on_hover_only(editor_bot, qtbot):
    """Test that the slider range indicator is visible only when hovering
    over the scrollflag area when the editor vertical scrollbar is visible.
    The scrollflag area should remain hidden at all times when the editor
    vertical scrollbar is not visible."""
    editor = editor_bot
    sfa = editor.scrollflagarea
    editor.show()

    # Set a short text in the editor and assert that the slider is not visible.
    editor.set_text(short_code)
    editor.resize(450, 150)
    qtbot.waitUntil(lambda: not sfa.slider)

    # Move the mouse cursor to the center of the scrollflagarea and assert
    # that the slider range indicator remains hidden. The slider range
    # indicator should remains hidden at all times when the vertical scrollbar
    # of the editor is not visible.
    x = int(sfa.width()/2)
    y = int(sfa.height()/2)
    qtbot.mouseMove(sfa, pos=QPoint(x, y), delay=-1)

    assert sfa._range_indicator_is_visible is False

    # Set a long text in the editor and assert that the slider is visible.
    editor.set_text(long_code)
    editor.resize(450, 150)
    qtbot.waitUntil(lambda: sfa.slider)

    # Move the mouse cursor to the center of the scrollflagarea and assert
    # that the slider range indicator is now shown. When the vertical scrollbar
    # of the editor is visible, the slider range indicator should be visible
    # only when the mouse cursor hover above the scrollflagarea.
    x = int(sfa.width()/2)
    y = int(sfa.height()/2)
    qtbot.mouseMove(sfa, pos=QPoint(x, y), delay=-1)
    qtbot.wait(500)

    assert sfa._range_indicator_is_visible is True

    # Move the mouse cursor outside of the scrollflagarea and assert that the
    # slider range indicator becomes hidden.
    x = int(editor.width()/2)
    y = int(editor.height()/2)
    qtbot.mouseMove(editor, pos=QPoint(x, y), delay=-1)
    qtbot.waitUntil(lambda: not sfa._range_indicator_is_visible)


def test_range_indicator_alt_modifier_response(editor_bot, qtbot):
    """Test that the slider range indicator is visible while the alt key is
    held down while the cursor is over the editor, but outside of the
    scrollflag area. In addition, while the alt key is held down, mouse
    click events in the editor should be forwarded to the scrollfag area and
    should set the value of the editor vertical scrollbar."""
    editor = editor_bot
    sfa = editor.scrollflagarea
    sfa._unit_testing = True
    vsb = editor.verticalScrollBar()

    editor.show()
    editor.resize(600, 150)

    # Set a long text in the editor and assert that the slider is visible.
    editor.set_text(long_code)
    qtbot.waitUntil(lambda: sfa.slider)
    qtbot.wait(500)

    # Set the cursor position to the center of the editor.
    w = editor.width()
    h = editor.height()
    qtbot.mousePress(editor, Qt.LeftButton, pos=QPoint(w//2, h//2))

    # Hold the alt key and assert that the slider range indicator is visible.
    # Because it is not possible to simulate the action of holding the alt
    # key down in pytest-qt, this is done through a flag in the ScrollFlagArea
    # that is set to True when pressing the alt key and to false when releasing
    # it. This flag is only used for testing purpose.
    qtbot.keyPress(editor, Qt.Key_Alt)
    editor.resize(600, 150)
    x = int(sfa.width()/2)
    y = int(sfa.height()/2)
    qtbot.mouseMove(sfa, pos=QPoint(x, y), delay=-1)
    qtbot.waitUntil(lambda: sfa._range_indicator_is_visible)

    # While the alt key is pressed, click with the mouse in the middle of the
    # editor's height and assert that the editor vertical scrollbar has moved
    # to its middle range position.
    with qtbot.waitSignal(editor.sig_alt_left_mouse_pressed, raising=True):
        qtbot.mousePress(editor.viewport(), Qt.LeftButton,
                         Qt.AltModifier, QPoint(w//2, h//2))
    assert vsb.value() == (vsb.minimum()+vsb.maximum())//2

    # While the alt key is pressed, click with the mouse at the top of the
    # editor's height and assert that the editor vertical scrollbar has moved
    # to its minimum position.
    with qtbot.waitSignal(editor.sig_alt_left_mouse_pressed, raising=True):
        qtbot.mousePress(editor.viewport(), Qt.LeftButton,
                         Qt.AltModifier, QPoint(w//2, 1))
    assert vsb.value() == vsb.minimum()

    # While the alt key is pressed, click with the mouse at the bottom of the
    # editor's height and assert that the editor vertical scrollbar has moved
    # to its maximum position.
    with qtbot.waitSignal(editor.sig_alt_left_mouse_pressed, raising=True):
        qtbot.mousePress(editor.viewport(), Qt.LeftButton,
                         Qt.AltModifier, QPoint(w//2, h-1))
    assert vsb.value() == vsb.maximum()

    # Release the alt key and assert that the slider range indicator is
    # not visible.
    editor.resize(600, 150)
    x = int(sfa.width()/2)
    y = int(sfa.height()/2)
    qtbot.mouseMove(sfa, pos=QPoint(x*100, y), delay=-1)
    qtbot.keyRelease(editor, Qt.Key_Alt)
    qtbot.waitUntil(lambda: not sfa._range_indicator_is_visible, timeout=3000)


if __name__ == "__main__":  # pragma: no cover
    pytest.main([os.path.basename(__file__)])



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""Tests for the report error dialog."""

from unittest.mock import Mock, MagicMock

# Third party imports
import pytest
from qtpy.QtCore import Qt, QUrl

# Local imports
from spyder import __project_url__
from spyder.widgets.reporterror import (DESC_MIN_CHARS, TITLE_MIN_CHARS,
                                        SpyderErrorDialog)

# =============================================================================
# Fixtures
# =============================================================================
@pytest.fixture
def error_dialog(qtbot):
    """Set up error report dialog."""
    widget = SpyderErrorDialog(None)
    qtbot.addWidget(widget)
    widget.show()
    return widget


# =============================================================================
# Tests
# =============================================================================
def test_dialog(error_dialog, qtbot):
    """Test that error report dialog UI behaves properly."""
    dlg = error_dialog
    desc_text = "1" * DESC_MIN_CHARS
    title_text = "1" * TITLE_MIN_CHARS

    # Assert Submit button is disabled at first
    assert not dlg.submit_btn.isEnabled()

    # Introduce MIN_CHARS to input_description
    qtbot.keyClicks(dlg.input_description, desc_text)
    qtbot.keyClicks(dlg.title, title_text)

    # Assert Submit button is now enabled
    assert dlg.submit_btn.isEnabled()

    # Assert cut leaves the header
    dlg.input_description.selectAll()
    dlg.input_description.cut()
    assert dlg.input_description.toPlainText() == dlg.input_description.header

    # Assert delete leaves the header
    qtbot.keyClicks(dlg.input_description, desc_text)
    dlg.input_description.selectAll()
    dlg.input_description.delete()
    assert dlg.input_description.toPlainText() == dlg.input_description.header

    # Assert delete doesn't remove characters on the header
    ini_pos = dlg.input_description.get_position('sof')
    dlg.input_description.set_cursor_position(ini_pos)
    dlg.input_description.delete()
    assert dlg.input_description.toPlainText() == dlg.input_description.header

    # Assert backspace works as expected
    qtbot.keyClicks(dlg.input_description, desc_text)
    qtbot.keyPress(dlg.input_description, Qt.Key_Backspace)
    assert not dlg.submit_btn.isEnabled()

    dlg.input_description.selectAll()
    qtbot.keyPress(dlg.input_description, Qt.Key_Backspace)
    assert dlg.input_description.toPlainText() == dlg.input_description.header

    ini_pos = dlg.input_description.get_position('sof')
    dlg.input_description.set_cursor_position(ini_pos)
    dlg.input_description.set_cursor_position('eol')
    qtbot.keyPress(dlg.input_description, Qt.Key_Backspace)
    assert dlg.input_description.toPlainText() == dlg.input_description.header

    # Assert chars label works as expected
    assert dlg.desc_chars_label.text() == '{} more characters to go...'.format(DESC_MIN_CHARS)
    qtbot.keyClicks(dlg.input_description, desc_text)
    assert dlg.desc_chars_label.text() == 'Description complete; thanks!'


def test_report_issue_url(monkeypatch):
    """Test that report_issue sends the data, and to correct url."""
    body = 'This is an example error report body text.'
    title = 'Uncreative issue title here'
    target_url_base = __project_url__ + '/issues/new'

    MockQDesktopServices = MagicMock()
    attr_to_patch = ('spyder.widgets.reporterror.QDesktopServices')
    monkeypatch.setattr(attr_to_patch, MockQDesktopServices)

    # Test when body != None, i.e. when auto-submitting error to Github
    target_url = QUrl(target_url_base + '?body=' + body)
    SpyderErrorDialog.open_web_report(body=body, title=None)
    assert MockQDesktopServices.openUrl.call_count == 1
    MockQDesktopServices.openUrl.assert_called_with(target_url)

    # Test when body != None and title != None
    target_url = QUrl(target_url_base + '?body=' + body
                      + "&title=" + title)
    SpyderErrorDialog.open_web_report(body=body, title=title)
    assert MockQDesktopServices.openUrl.call_count == 2
    MockQDesktopServices.openUrl.assert_called_with(target_url)


def test_render_issue():
    """Test that render issue works without errors and returns text."""
    test_description = "This is a test description"
    test_traceback = "An error occurred. Oh no!"

    # Test when description and traceback are not provided
    test_issue_1 = SpyderErrorDialog.render_issue()
    assert type(test_issue_1) == str
    assert len(test_issue_1) > 100

    # Test when description and traceback are provided
    test_issue_2 = SpyderErrorDialog.render_issue(
        test_description, test_traceback)
    assert type(test_issue_2) == str
    assert len(test_issue_2) > 100
    assert test_description in test_issue_2
    assert test_traceback in test_issue_2


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Editor Plugin"""

import logging
import sys

from qtpy.QtCore import Signal

from spyder.api.fonts import SpyderFontType
from spyder.api.plugins import SpyderDockablePlugin, Plugins
from spyder.api.plugin_registration.decorators import (
    on_plugin_available,
    on_plugin_teardown,
)
from spyder.api.translations import _
from spyder.plugins.editor.api.run import (
    SelectionContextModificator,
    ExtraAction
)
from spyder.plugins.editor.confpage import EditorConfigPage
from spyder.plugins.editor.widgets.main_widget import (
    EditorMainWidget,
    EditorWidgetActions
)
from spyder.plugins.mainmenu.api import (
    ApplicationMenus,
    EditMenuSections,
    FileMenuSections,
    SearchMenuSections,
    SourceMenuSections
)
from spyder.plugins.toolbar.api import ApplicationToolbars
from spyder.plugins.run.api import RunContext


logger = logging.getLogger(__name__)


class Editor(SpyderDockablePlugin):
    """
    Editor plugin.
    """

    NAME = 'editor'
    REQUIRES = [Plugins.Console, Plugins.Preferences]
    OPTIONAL = [
        Plugins.Completions,
        Plugins.Debugger,
        Plugins.IPythonConsole,
        Plugins.MainMenu,
        Plugins.Projects,
        Plugins.OutlineExplorer,
        Plugins.Run,
        Plugins.StatusBar,
        Plugins.Switcher,
        Plugins.Toolbar
    ]
    WIDGET_CLASS = EditorMainWidget
    CONF_SECTION = NAME
    CONF_WIDGET_CLASS = EditorConfigPage
    CONF_FILE = False

    # ---- Signals
    # ------------------------------------------------------------------------
    sig_dir_opened = Signal(str)
    """
    This signal is emitted when the editor changes the current directory.

    Parameters
    ----------
    new_working_directory: str
        The new working directory path.

    Notes
    -----
    This option is available on the options menu of the editor plugin
    """

    sig_file_opened_closed_or_updated = Signal(str, str)
    """
    This signal is emitted when a file is opened, closed or updated,
    including switching among files.

    Parameters
    ----------
    filename: str
        Name of the file that was opened, closed or updated.
    language: str
        Name of the programming language of the file that was opened,
        closed or updated.
    """

    # This signal is fired for any focus change among all editor stacks
    sig_editor_focus_changed = Signal()

    sig_help_requested = Signal(dict)
    """
    This signal is emitted to request help on a given object `name`.

    Parameters
    ----------
    help_data: dict
        Dictionary required by the Help pane to render a docstring.

    Examples
    --------
    >>> help_data = {
        'obj_text': str,
        'name': str,
        'argspec': str,
        'note': str,
        'docstring': str,
        'force_refresh': bool,
        'path': str,
    }

    See Also
    --------
    :py:meth:spyder.plugins.editor.widgets.editorstack.EditorStack.send_to_help
    """

    sig_open_files_finished = Signal()
    """
    This signal is emitted when the editor finished to open files.
    """

    sig_codeeditor_created = Signal(object)
    """
    This signal is emitted when a codeeditor is created.

    Parameters
    ----------
    codeeditor: spyder.plugins.editor.widgets.codeeditor.CodeEditor
        The codeeditor.
    """

    sig_codeeditor_deleted = Signal(object)
    """
    This signal is emitted when a codeeditor is closed.

    Parameters
    ----------
    codeeditor: spyder.plugins.editor.widgets.codeeditor.CodeEditor
        The codeeditor.
    """

    sig_codeeditor_changed = Signal(object)
    """
    This signal is emitted when the current codeeditor changes.

    Parameters
    ----------
    codeeditor: spyder.plugins.editor.widgets.codeeditor.CodeEditor
        The codeeditor.
    """

    # ---- SpyderDockablePlugin API
    # ------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return _('Editor')

    @staticmethod
    def get_description():
        return _(
            "Edit Python, Markdown, Cython and many other types of text files."
        )

    @classmethod
    def get_icon(cls):
        return cls.create_icon('edit')

    def on_initialize(self):
        widget = self.get_widget()

        # ---- Help related signals
        widget.sig_help_requested.connect(self.sig_help_requested)

        # ---- General signals
        widget.starting_long_process.connect(self.before_long_process)
        widget.ending_long_process.connect(self.after_long_process)
        widget.sig_dir_opened.connect(self.sig_dir_opened)
        widget.sig_file_opened_closed_or_updated.connect(
            self.sig_file_opened_closed_or_updated
        )
        widget.sig_open_files_finished.connect(self.sig_open_files_finished)

        # ---- CodeEditor related signals
        widget.sig_codeeditor_created.connect(self.sig_codeeditor_created)
        widget.sig_codeeditor_deleted.connect(self.sig_codeeditor_deleted)
        widget.sig_codeeditor_changed.connect(self.sig_codeeditor_changed)
        widget.sig_editor_focus_changed.connect(self.sig_editor_focus_changed)

        # ---- Plugin related signals
        widget.sig_switch_to_plugin_requested.connect(
            lambda: self.switch_to_plugin(force_focus=True)
        )

    @on_plugin_available(plugin=Plugins.Preferences)
    def on_preferences_available(self):
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.register_plugin_preferences(self)

    @on_plugin_teardown(plugin=Plugins.Preferences)
    def on_preferences_teardown(self):
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.deregister_plugin_preferences(self)

    @on_plugin_available(plugin=Plugins.StatusBar)
    def on_statusbar_available(self):
        # Add status widgets
        statusbar = self.get_plugin(Plugins.StatusBar)
        widget = self.get_widget()
        statusbar.add_status_widget(widget.readwrite_status)
        statusbar.add_status_widget(widget.eol_status)
        statusbar.add_status_widget(widget.encoding_status)
        statusbar.add_status_widget(widget.cursorpos_status)
        statusbar.add_status_widget(widget.vcs_status)

    @on_plugin_teardown(plugin=Plugins.StatusBar)
    def on_statusbar_teardown(self):
        # Remove status widgets
        statusbar = self.get_plugin(Plugins.StatusBar)
        widget = self.get_widget()
        statusbar.remove_status_widget(widget.readwrite_status.ID)
        statusbar.remove_status_widget(widget.eol_status.ID)
        statusbar.remove_status_widget(widget.encoding_status.ID)
        statusbar.remove_status_widget(widget.cursorpos_status.ID)
        statusbar.remove_status_widget(widget.vcs_status.ID)

    @on_plugin_available(plugin=Plugins.Run)
    def on_run_available(self):
        widget = self.get_widget()
        run = self.get_plugin(Plugins.Run)

        widget.sig_editor_focus_changed_uuid.connect(
            run.switch_focused_run_configuration
        )
        widget.sig_register_run_configuration_provider_requested.connect(
            lambda supported_extensions:
                run.register_run_configuration_provider(
                    self.NAME, supported_extensions
                )
        )
        widget.sig_deregister_run_configuration_provider_requested.connect(
            lambda unsupported_extensions:
                run.deregister_run_configuration_provider(
                    self.NAME, unsupported_extensions
                )
        )

        # This is necessary to register run configs that were added before Run
        # is available
        for extension in widget.supported_run_extensions:
            run.register_run_configuration_provider(self.NAME, [extension])

        # Buttons creation
        run.create_run_button(
            RunContext.Cell,
            _("Run cell"),
            icon=self.create_icon('run_cell'),
            tip=_("Run cell"),
            shortcut_context=self.NAME,
            register_shortcut=True,
            add_to_toolbar=True,
            add_to_menu=True
        )
        run.create_run_button(
            RunContext.Cell,
            _("Run cell and advance"),
            icon=self.create_icon('run_cell_advance'),
            tip=_("Run cell and advance"),
            shortcut_context=self.NAME,
            register_shortcut=True,
            add_to_toolbar=True,
            add_to_menu=True,
            extra_action_name=ExtraAction.Advance
        )
        run.create_run_button(
            RunContext.Cell,
            _("Re-run last cell"),
            tip=_("Re run last cell "),
            shortcut_context=self.NAME,
            register_shortcut=True,
            add_to_menu=True,
            re_run=True
        )
        run.create_run_button(
            RunContext.Selection,
            _("Run &selection or current line"),
            icon=self.create_icon('run_selection'),
            tip=_("Run selection or current line"),
            shortcut_context=self.NAME,
            register_shortcut=True,
            add_to_toolbar=True,
            add_to_menu=True,
            extra_action_name=ExtraAction.Advance,
        )
        run.create_run_button(
            RunContext.Selection,
            _("Run &to line"),
            tip=_("Run selection up to the current line"),
            shortcut_context=self.NAME,
            register_shortcut=True,
            add_to_toolbar=False,
            add_to_menu=True,
            context_modificator=SelectionContextModificator.ToLine
        )
        run.create_run_button(
            RunContext.Selection,
            _("Run &from line"),
            tip=_("Run selection from the current line"),
            shortcut_context=self.NAME,
            register_shortcut=True,
            add_to_toolbar=False,
            add_to_menu=True,
            context_modificator=SelectionContextModificator.FromLine
        )

    @on_plugin_teardown(plugin=Plugins.Run)
    def on_run_teardown(self):
        widget = self.get_widget()
        run = self.get_plugin(Plugins.Run)
        run.deregister_run_configuration_provider(
            self.NAME, widget.supported_run_extensions
        )
        run.destroy_run_button(RunContext.Cell)
        run.destroy_run_button(
            RunContext.Cell,
            extra_action_name=ExtraAction.Advance
        )
        run.destroy_run_button(RunContext.Cell, re_run=True)
        run.destroy_run_button(
            RunContext.Selection,
            extra_action_name=ExtraAction.Advance
        )
        run.destroy_run_button(
            RunContext.Selection,
            context_modificator=SelectionContextModificator.ToLine
        )
        run.destroy_run_button(
            RunContext.Selection,
            context_modificator=SelectionContextModificator.FromLine
        )

    @on_plugin_available(plugin=Plugins.MainMenu)
    def on_mainmenu_available(self):
        widget = self.get_widget()
        mainmenu = self.get_plugin(Plugins.MainMenu)
        # ---- File menu ----
        # Print
        print_actions = [
            widget.print_preview_action,
            widget.print_action,
        ]
        for print_action in print_actions:
            mainmenu.add_item_to_application_menu(
                print_action,
                menu_id=ApplicationMenus.File,
                section=FileMenuSections.Print,
                before_section=FileMenuSections.Close
            )

        # Close
        close_actions = [
            widget.close_action,
            widget.close_all_action
        ]
        for close_action in close_actions:
            mainmenu.add_item_to_application_menu(
                close_action,
                menu_id=ApplicationMenus.File,
                section=FileMenuSections.Close,
                before_section=FileMenuSections.Restart
            )

        # Navigation
        if sys.platform == 'darwin':
            tab_navigation_actions = [
                widget.go_to_previous_file_action,
                widget.go_to_next_file_action
            ]
            for tab_navigation_action in tab_navigation_actions:
                mainmenu.add_item_to_application_menu(
                    tab_navigation_action,
                    menu_id=ApplicationMenus.File,
                    section=FileMenuSections.Navigation,
                    before_section=FileMenuSections.Restart
                )

        # Open section
        open_actions = [
            widget.open_action,
            widget.open_last_closed_action,
            widget.recent_file_menu,
        ]
        for open_action in open_actions:
            mainmenu.add_item_to_application_menu(
                open_action,
                menu_id=ApplicationMenus.File,
                section=FileMenuSections.Open,
                before_section=FileMenuSections.Save
            )

        # Save section
        save_actions = [
            widget.save_action,
            widget.save_all_action,
            widget.save_as_action,
            widget.save_copy_as_action,
            widget.revert_action,
        ]
        for save_action in save_actions:
            mainmenu.add_item_to_application_menu(
                save_action,
                menu_id=ApplicationMenus.File,
                section=FileMenuSections.Save,
                before_section=FileMenuSections.Print
            )

        # New Section
        mainmenu.add_item_to_application_menu(
            widget.new_action,
            menu_id=ApplicationMenus.File,
            section=FileMenuSections.New,
            before_section=FileMenuSections.Open
        )

        # ---- Edit menu ----
        edit_menu = mainmenu.get_application_menu(ApplicationMenus.Edit)
        edit_menu.aboutToShow.connect(widget.update_edit_menu)

        # UndoRedo section
        for action in [widget.undo_action, widget.redo_action]:
            mainmenu.add_item_to_application_menu(
                action,
                menu_id=ApplicationMenus.Edit,
                section=EditMenuSections.UndoRedo,
                before_section=EditMenuSections.Editor
            )

        # Copy section
        for action in [
                widget.cut_action, widget.copy_action, widget.paste_action,
                widget.selectall_action]:
            mainmenu.add_item_to_application_menu(
                action,
                menu_id=ApplicationMenus.Edit,
                section=EditMenuSections.Copy,
                before_section=EditMenuSections.Editor
            )

        # Editor section
        for edit_item in widget.edit_menu_actions:
            mainmenu.add_item_to_application_menu(
                edit_item,
                menu_id=ApplicationMenus.Edit,
                section=EditMenuSections.Editor
            )

        # ---- Search menu ----
        search_menu = mainmenu.get_application_menu(ApplicationMenus.Search)
        search_menu.aboutToShow.connect(widget.update_search_menu)

        for search_item in widget.search_menu_actions:
            mainmenu.add_item_to_application_menu(
                search_item,
                menu_id=ApplicationMenus.Search,
                section=SearchMenuSections.FindInText,
                before_section=SearchMenuSections.FindInFiles
            )

        # ---- Source menu ----
        source_menu = mainmenu.get_application_menu(
            ApplicationMenus.Source
        )
        source_menu.aboutToShow.connect(widget.refresh_formatter_name)

        # Cursor section
        source_menu_cursor_actions = [
            widget.previous_edit_cursor_action,
            widget.previous_cursor_action,
            widget.next_cursor_action,
        ]
        for cursor_item in source_menu_cursor_actions:
            mainmenu.add_item_to_application_menu(
                cursor_item,
                menu_id=ApplicationMenus.Source,
                section=SourceMenuSections.Cursor,
                before_section=SourceMenuSections.Formatting
            )

        # Formatting section
        source_menu_formatting_actions = [
            widget.eol_menu,
            widget.trailingspaces_action,
            widget.fixindentation_action,
            widget.formatting_action
        ]
        for formatting_item in source_menu_formatting_actions:
            mainmenu.add_item_to_application_menu(
                formatting_item,
                menu_id=ApplicationMenus.Source,
                section=SourceMenuSections.Formatting,
                before_section=SourceMenuSections.CodeAnalysis
            )

        # Options section
        source_menu_option_actions = widget.checkable_actions.values()
        for option_item in source_menu_option_actions:
            mainmenu.add_item_to_application_menu(
                option_item,
                menu_id=ApplicationMenus.Source,
                section=SourceMenuSections.Options,
                before_section=SourceMenuSections.Linting
            )

        # Linting section
        source_menu_linting_actions = [
            widget.todo_list_action,
            widget.warning_list_action,
            widget.previous_warning_action,
            widget.next_warning_action,
        ]
        for linting_item in source_menu_linting_actions:
            mainmenu.add_item_to_application_menu(
                linting_item,
                menu_id=ApplicationMenus.Source,
                section=SourceMenuSections.Linting,
                before_section=SourceMenuSections.Cursor
            )

    @on_plugin_teardown(plugin=Plugins.MainMenu)
    def on_mainmenu_teardown(self):
        widget = self.get_widget()
        mainmenu = self.get_plugin(Plugins.MainMenu)
        # ---- File menu ----
        # Print
        print_actions = [
            widget.print_preview_action,
            widget.print_action,
        ]
        for print_action in print_actions:
            mainmenu.remove_item_from_application_menu(
                print_action,
                menu_id=ApplicationMenus.File
            )

        # Close
        close_actions = [
            widget.close_action,
            widget.close_all_action
        ]
        for close_action in close_actions:
            mainmenu.remove_item_from_application_menu(
                close_action,
                menu_id=ApplicationMenus.File
            )

        # Navigation
        if sys.platform == 'darwin':
            tab_navigation_actions = [
                widget.go_to_previous_file_action,
                widget.go_to_next_file_action
            ]
            for tab_navigation_action in tab_navigation_actions:
                mainmenu.remove_item_from_application_menu(
                    tab_navigation_action,
                    menu_id=ApplicationMenus.File
                )

        # Open section
        open_actions = [
            widget.open_action,
            widget.open_last_closed_action,
            widget.recent_file_menu,
        ]
        for open_action in open_actions:
            mainmenu.remove_item_from_application_menu(
                open_action,
                menu_id=ApplicationMenus.File
            )

        # Save section
        save_actions = [
            widget.save_action,
            widget.save_all_action,
            widget.save_as_action,
            widget.save_copy_as_action,
            widget.revert_action,
        ]
        for save_action in save_actions:
            mainmenu.remove_item_from_application_menu(
                save_action,
                menu_id=ApplicationMenus.File
            )

        # New Section
        mainmenu.remove_item_from_application_menu(
            widget.new_action,
            menu_id=ApplicationMenus.File
        )

        # ---- Edit menu ----
        edit_menu = mainmenu.get_application_menu(ApplicationMenus.Edit)
        edit_menu.aboutToShow.disconnect(widget.update_edit_menu)

        # UndoRedo section
        for action in [widget.undo_action, widget.redo_action]:
            mainmenu.remove_item_from_application_menu(
                action,
                menu_id=ApplicationMenus.Edit
            )

        # Copy section
        for action in [
                widget.cut_action, widget.copy_action, widget.paste_action,
                widget.selectall_action]:
            mainmenu.remove_item_from_application_menu(
                action,
                menu_id=ApplicationMenus.Edit
            )

        # Editor section
        for edit_item in widget.edit_menu_actions:
            mainmenu.remove_item_from_application_menu(
                edit_item,
                menu_id=ApplicationMenus.Edit
            )

        # ---- Search menu ----
        search_menu = mainmenu.get_application_menu(ApplicationMenus.Search)
        search_menu.aboutToShow.disconnect(widget.update_search_menu)

        for search_item in widget.search_menu_actions:
            mainmenu.remove_item_from_application_menu(
                search_item,
                menu_id=ApplicationMenus.Search
            )

        # ---- Source menu ----
        source_menu = mainmenu.get_application_menu(
            ApplicationMenus.Source
        )
        source_menu.aboutToShow.disconnect(widget.refresh_formatter_name)

        # Cursor section
        source_menu_cursor_actions = [
            widget.previous_edit_cursor_action,
            widget.previous_cursor_action,
            widget.next_cursor_action,
        ]
        for cursor_item in source_menu_cursor_actions:
            mainmenu.remove_item_from_application_menu(
                cursor_item,
                menu_id=ApplicationMenus.Source
            )

        # Formatting section
        source_menu_formatting_actions = [
            widget.eol_menu,
            widget.trailingspaces_action,
            widget.fixindentation_action,
            widget.formatting_action
        ]
        for formatting_item in source_menu_formatting_actions:
            mainmenu.remove_item_from_application_menu(
                formatting_item,
                menu_id=ApplicationMenus.Source
            )

        # Options section
        source_menu_option_actions = widget.checkable_actions.values()
        for option_item in source_menu_option_actions:
            mainmenu.remove_item_from_application_menu(
                option_item,
                menu_id=ApplicationMenus.Source
            )

        # Linting section
        source_menu_linting_actions = [
            widget.todo_list_action,
            widget.warning_list_action,
            widget.previous_warning_action,
            widget.next_warning_action,
        ]
        for linting_item in source_menu_linting_actions:
            mainmenu.remove_item_from_application_menu(
                linting_item,
                menu_id=ApplicationMenus.Source
            )

    @on_plugin_available(plugin=Plugins.Toolbar)
    def on_toolbar_available(self):
        widget = self.get_widget()
        toolbar = self.get_plugin(Plugins.Toolbar)
        file_toolbar_actions = [
            widget.new_action,
            widget.open_action,
            widget.save_action,
            widget.save_all_action,
            widget.create_new_cell
        ]
        for file_toolbar_action in file_toolbar_actions:
            toolbar.add_item_to_application_toolbar(
                file_toolbar_action,
                toolbar_id=ApplicationToolbars.File,
            )

    @on_plugin_teardown(plugin=Plugins.Toolbar)
    def on_toolbar_teardown(self):
        toolbar = self.get_plugin(Plugins.Toolbar)
        file_toolbar_actions = [
            EditorWidgetActions.NewFile,
            EditorWidgetActions.OpenFile,
            EditorWidgetActions.SaveFile,
            EditorWidgetActions.SaveAll,
            EditorWidgetActions.NewCell
        ]
        for file_toolbar_action_id in file_toolbar_actions:
            toolbar.remove_item_from_application_toolbar(
                file_toolbar_action_id,
                toolbar_id=ApplicationToolbars.File,
            )

    @on_plugin_available(plugin=Plugins.Completions)
    def on_completions_available(self):
        widget = self.get_widget()
        completions = self.get_plugin(Plugins.Completions)

        widget.sig_after_configuration_update_requested.connect(
            completions.after_configuration_update
        )
        self.sig_file_opened_closed_or_updated.connect(
            completions.file_opened_closed_or_updated
        )

        completions.sig_language_completions_available.connect(
            widget.register_completion_capabilities)
        completions.sig_open_file.connect(widget.load)
        completions.sig_stop_completions.connect(
            widget.stop_completion_services)

    @on_plugin_teardown(plugin=Plugins.Completions)
    def on_completions_teardown(self):
        widget = self.get_widget()
        completions = self.get_plugin(Plugins.Completions)

        widget.sig_after_configuration_update_requested.disconnect(
            completions.after_configuration_update
        )
        self.sig_file_opened_closed_or_updated.disconnect(
            completions.file_opened_closed_or_updated
        )

        completions.sig_language_completions_available.disconnect(
            widget.register_completion_capabilities)
        completions.sig_open_file.disconnect(widget.load)
        completions.sig_stop_completions.disconnect(
            widget.stop_completion_services)

    @on_plugin_available(plugin=Plugins.OutlineExplorer)
    def on_outlineexplorer_available(self):
        widget = self.get_widget()
        outline = self.get_plugin(Plugins.OutlineExplorer)
        outline_widget = outline.get_widget()

        widget.set_outlineexplorer(outline_widget)

    @on_plugin_teardown(plugin=Plugins.OutlineExplorer)
    def on_outlinexplorer_teardown(self):
        self.get_widget().set_outlineexplorer(None)

    @on_plugin_available(plugin=Plugins.IPythonConsole)
    def on_ipyconsole_available(self):
        widget = self.get_widget()
        ipyconsole = self.get_plugin(Plugins.IPythonConsole)

        ipyconsole.register_spyder_kernel_call_handler(
            'cell_count', widget.handle_cell_count
        )
        ipyconsole.register_spyder_kernel_call_handler(
            'current_filename', widget.handle_current_filename
        )
        ipyconsole.register_spyder_kernel_call_handler(
            'get_file_code', widget.handle_get_file_code
        )
        ipyconsole.register_spyder_kernel_call_handler(
            'run_cell', widget.handle_run_cell
        )

    @on_plugin_teardown(plugin=Plugins.IPythonConsole)
    def on_ipyconsole_teardown(self):
        ipyconsole = self.get_plugin(Plugins.IPythonConsole)

        ipyconsole.unregister_spyder_kernel_call_handler('cell_count')
        ipyconsole.unregister_spyder_kernel_call_handler('current_filename')
        ipyconsole.unregister_spyder_kernel_call_handler('get_file_code')
        ipyconsole.unregister_spyder_kernel_call_handler('run_cell')

    @on_plugin_available(plugin=Plugins.Switcher)
    def on_switcher_available(self):
        switcher = self.get_plugin(Plugins.Switcher)
        self.get_widget().set_switcher(switcher)

    @on_plugin_teardown(plugin=Plugins.Switcher)
    def on_switcher_teardown(self):
        self.get_widget().set_switcher(None)

    @on_plugin_available(plugin=Plugins.Projects)
    def on_projects_available(self):
        projects = self.get_plugin(Plugins.Projects)
        projects.sig_project_loaded.connect(self._on_project_loaded)
        projects.sig_project_closed.connect(self._on_project_closed)

    @on_plugin_teardown(plugin=Plugins.Projects)
    def on_projects_teardown(self):
        projects = self.get_plugin(Plugins.Projects)
        projects.sig_project_loaded.disconnect(self._on_project_loaded)
        projects.sig_project_closed.disconnect(self._on_project_closed)

    def update_font(self):
        """Update font from Preferences"""
        font = self.get_font(SpyderFontType.Monospace)
        self.get_widget().update_font(font)

    def before_mainwindow_visible(self):
        # Don't move this to on_mainwindow_visible because the window appears
        # empty while the recovery dialog is shown.
        self.get_widget().autosave.try_recover_from_autosave()

    def on_mainwindow_visible(self):
        widget = self.get_widget()
        widget.restore_scrollbar_position()

    def can_close(self):
        editorstack = self.get_widget().editorstacks[0]
        return editorstack.save_if_changed(cancelable=True)

    def on_close(self, cancelable=False):
        widget = self.get_widget()

        if not self.get_widget().get_active_project_path():
            filenames = widget.get_filenames()
            self.set_conf('filenames', filenames)

    # ---- Public API
    # ------------------------------------------------------------------------
    def get_codeeditor_for_filename(self, filename):
        """
        Get `CodeEditor` instance associated with the given filename.

        Parameters
        ----------
        filename : str
            File path associated with a CodeEditor instance.

        Returns
        -------
        spyder.plugins.editor.codeeditor.CodeEditor
            `CodeEditor` associated with the given filename.
        """
        return self.get_widget().get_editor(filename)

    def refresh(self):
        """
        Refresh main widget.
        """
        self.get_widget().refresh()

    def load(self, *args, **kwargs):
        """
        Load a file or a group of files.

        Parameters
        ----------
        filenames: Optional[list]
            Filenames to load.
        goto: Optional[int]
            If goto is not None, it represents a line to go to. Used alongside
            `start_column` and `end_column`. Alternatively, the first match of
            `word` is used as a position.
        word: Optional[str]
            The `word` to use to set the cursor position when using `goto`.
        editorwindow: Optional[spyder.plugins.editor.widgets.window.EditorMainWindow]  # noqa
            Load in the given editorwindow (useful when clicking in the Outline
            explorer with multiple editor windows).
        processevents: Optional[bool]
            Determines if `processEvents()` should be called at the end of this
            method (set to `False` to prevent keyboard events from creeping
            through to the editor while debugging).
        start_column: Optional[int]
            The start position in the line (goto)
        end_column: Optional[int]
            The length (so that the end position is `start_column` +
            `end_column`), when providing a `goto` line.
        set_focus: Optional[bool]
            If the opened file should gain focus. `True` by default.
        add_where: Optional[str]
            Position where to add the new file finfo (affects the files tab
            order). Possible values are: `start` to make the file the first and
             `end` (or any other value) to append.
        """
        return self.get_widget().load(*args, **kwargs)

    def load_edit(self, filename):
        """
        Load a `filename` passing to the base `load` method the `main_widget`
        as the `editorwindow` to force focus.

        Used by `spyder.plugins.outlineexplorer.plugin.[on_editor_available|on_editor_teardown]`  # noqa

        Parameters
        ----------
        filename: str
            Filename to load.
        """
        widget = self.get_widget()
        return self.get_widget().load(filenames=filename, editorwindow=widget)

    def load_edit_goto(self, filename, goto, word):
        """
        Load a `filename` and put the cursor in the line given by `goto` and
        `word`, passing to the `load` call the `main_widget` as the
        `editorwindow` to force focus.

        Used by `spyder.plugins.outlineexplorer.plugin.[on_editor_available|on_editor_teardown]`  # noqa

        Parameters
        ----------
        filename: str
            Filename to load.
        goto: int
            Represents a line to go to.
        word: str
            The `word` to use to set the cursor position when using `goto`.
        """
        widget = self.get_widget()
        return widget.load(
            filenames=filename, goto=goto, word=word, editorwindow=widget
        )

    def new(self, *args, **kwargs):
        """
        Create a new file.

        Parameters
        ----------
        fname: Optional[str]
            Name of the file to be created. The default is `None`.
            If `None`, `fname` will be named `untitledXX.py`. No actual file
            will be created until it is saved manually by the user.
        editorstack: Optional[spyder.plugins.editor.widgets.editorstack.EditorStack]  # noqa
            Reference to the `EditorStack` instance that will be used to:
                * Get `untitledXX.py` numbering for the file name.
                * Check if a file with the same name already exists and it is
                  closeable.
                * Set file as the current focused file.
            The default is `None`. If that's the case, the current
            `EditorStack` is used. See the `get_current_editorstack` method for
            more details.
        text: Optional[str]
            Base text content that will be added to the file. The default is
            `None`. If that's the case, the default content created will be 
            created via a template file. See
            `Preferences > Editor > Advanced settings > Edit template for new files`  # noqa
        """
        return self.get_widget().new(*args, **kwargs)

    def removed(self, filename):
        """
        Close file given his filename since it was removed.

        It's used, for instance, when a file was removed in the File or Project
        explorer plugins.

        Parameters
        ----------
        filename: str
            File path to be closed/removed.
        """
        return self.get_widget().removed(filename)

    def removed_tree(self, dirname):
        """
        Close files given a directory since it was removed.

        It's used, for instance, when a directory was removed in the File or
        Project explorer plugins.

        Parameters
        ----------
        dirname: str
            Base directory path of the files to be closed/removed.
        """
        return self.get_widget().removed_tree(dirname)

    def renamed(self, *args, **kwargs):
        """
        Propagate file rename to editor stacks and autosave component.

        This method is called when a file is renamed in the File or Project
        explorer plugins. The file may not be opened in the editor.

        Parameters
        ----------
        source: str
            Initial filename path.
        dest: str
            New filename path.
        editorstack_id_str: Optional[str]
            The default is `None`. If not, the `EditorStack` instance whose
            identity corresponds to `editorstack_id_str` **doesn't** perform
            the file rename operation.
        """
        return self.get_widget().renamed(*args, **kwargs)

    def renamed_tree(self, *args, **kwargs):
        """
        Propagate directory rename to editor stacks and autosave component.

        This is used when the directory was renamed in File or Project explorer
        plugins.

        Parameters
        ----------
        source: str
            Initial directory path.
        dest: str
            New directory path.
        """
        return self.get_widget().renamed_tree(*args, **kwargs)

    def add_supported_run_configuration(self, *args, **kwargs):
        """
        Add a run configuration schema supported by the Editor.

        Parameters
        ----------
        config : spyder.plugins.editor.api.run.EditorRunConfiguration
            New run configuration schema to be added.
        """
        return self.get_widget().add_supported_run_configuration(
            *args, **kwargs
        )

    def remove_supported_run_configuration(self, *args, **kwargs):
        """
        Remove a run configuration schema supported by the Editor.

        Parameters
        ----------
        config : spyder.plugins.editor.api.run.EditorRunConfiguration
            Run configuration schema to be removed.
        """
        return self.get_widget().remove_supported_run_configuration(
            *args, **kwargs
        )

    def get_current_editor(self):
        """
        Get current `CodeEditor` instance if available.

        Returns
        -------
        spyder.plugins.editor.codeeditor.CodeEditor
            `CodeEditor` instance focused or available.
        """
        return self.get_widget().get_current_editor()

    def get_current_editorstack(self):
        """
        Get current `EditorStack` instance if available.

        Returns
        -------
        spyder.plugins.editor.editorstack.EditorStack
            `EditorStack` instance focused or available.
        """
        return self.get_widget().get_current_editorstack()

    def get_focus_widget(self):
        """
        Return the widget to give focus to.

        This happens when plugin's main widget is raised to the top-level.

        Returns
        -------
        spyder.plugins.editor.codeeditor.CodeEditor
            `CodeEditor` instance focused or available.
        """
        return self.get_widget().get_focus_widget()

    def setup_open_files(self,  close_previous_files=True):
        """
        Open the list of saved files per project.

        Also, open any files that the user selected in the recovery dialog and
        setup toolbars and menus for 'New window' instances (i.e. it calls the
        `setup_other_windows` method).

        Parameters
        ----------
        close_previous_files : Optional[bool]
            If any previously open file should be closed. Default `True`.
        """
        widget = self.get_widget()
        outline = self.get_plugin(Plugins.OutlineExplorer, error=False)
        if outline:
            widget.setup_other_windows(self._main, outline)
        return self.get_widget().setup_open_files(
            close_previous_files=close_previous_files
        )

    def save_open_files(self,):
        """Save the list of open files."""
        return self.get_widget().save_open_files()

    def save(self, index=None, force=False):
        """
        Save file.

        Parameters
        ----------
        index : Optional[int]
            Index related to the file position in the current editorstack.
            The default is `None`, which uses the current file index.
        force : Optional[bool]
            Force save regardless of file state. The default is `False`.

        Returns
        -------
        bool
            `True` if the save operation was sucessfull. `False` otherwise.
        """
        return self.get_widget().save(index=None, force=False)

    def save_bookmark(self, slot_num):
        """
        Save current line and position as bookmark.

        Parameters
        ----------
        slot_num : int
        """
        return self.get_widget().save_bookmark(slot_num)

    def load_bookmark(self, slot_num):
        """
        Set cursor to bookmarked file and position.

        Parameters
        ----------
        slot_num : int
        """
        return self.get_widget().load_bookmark(slot_num)

    def edit_template(self):
        """Edit `New file` template."""
        return self.get_widget().edit_template()

    def get_current_filename(self):
        """Get current editor 'filename'."""
        return self.get_widget().get_current_filename()

    def get_filenames(self):
        """
        Get list with all open files.

        Returns
        -------
        list
            A list with the names of all files currently opened in
            the editorstack.
        """
        return self.get_widget().get_filenames()

    def close_file(self):
        """Close current file."""
        return self.get_widget().close_file()

    def close_file_from_name(self, filename):
        """
        Close file from its name.

        Parameters
        ----------
        filename : str
            Filename to be closed.
        """
        return self.get_widget().close_file_from_name(filename)

    def close_all_files(self):
        """Close all opened files."""
        return self.get_widget().close_all_files()

    def go_to_line(self, line=None):
        """
        Open 'go to line' dialog.

        Parameters
        ----------
        line : Optional[int]
            Line to use for programatic calls without showing the dialog. The
            default is `None`.
        """
        return self.get_widget().go_to_line(line=line)

    def set_current_filename(self, *args, **kwargs):
        """
        Set current filename.

        Returns
        -------
        spyder.plugins.editor.codeeditor.CodeEditor
            The associated `CodeEditor` instance.
        """
        return self.get_widget().set_current_filename(*args, **kwargs)

    def set_current_project_path(self, root_path=None):
        """
        Set the current active project root path.

        Parameters
        ----------
        root_path: Optional[str]
            Path to current project root path. Default is `None`.
        """
        return self.get_widget().set_current_project_path(root_path=root_path)

    # ---- Private API
    # ------------------------------------------------------------------------
    # ---- Run related methods
    def _register_run_configuration_metadata(self, metadata):
        run = self.get_plugin(Plugins.Run, error=False)
        if run is not None:
            run.register_run_configuration_metadata(
                self.get_widget(), metadata
            )

    def _deregister_run_configuration_metadata(self, file_id):
        run = self.get_plugin(Plugins.Run, error=False)
        if run is not None:
            run.deregister_run_configuration_metadata(file_id)

    def _get_currently_selected_run_configuration(self):
        run = self.get_plugin(Plugins.Run, error=False)
        if run is not None:
            return run.get_currently_selected_configuration()

    def _switch_focused_run_configuration(self, file_id):
        run = self.get_plugin(Plugins.Run, error=False)
        if run is not None:
            run.switch_focused_run_configuration(file_id)

    # ---- Completions related methods
    def _register_file_completions(self, language, filename, codeeditor):
        completions = self.get_plugin(Plugins.Completions, error=False)
        status = None
        fallback_only = False
        if completions is not None:
            status = (
                completions.start_completion_services_for_language(
                    language.lower()
                )
            )
            completions.register_file(
                language.lower(), filename, codeeditor
            )
            fallback_only = completions.is_fallback_only(language.lower())
        return (status, fallback_only)

    def _send_completions_request(self, language, request, params):
        completions = self.get_plugin(Plugins.Completions, error=False)
        if completions is not None:
            completions.send_request(language, request, params)

    def _after_configuration_update(self, config):
        completions = self.get_plugin(Plugins.Completions, error=False)
        if completions is not None:
            completions.after_configuration_update(config)

    # ---- Projects related methods
    def _start_project_workspace_services(self):
        projects = self.get_plugin(Plugins.Projects, error=False)
        if projects is not None:
            projects.start_workspace_services()

    def _get_project_filenames(self):
        if self.get_widget().get_active_project_path():
            projects = self.get_plugin(Plugins.Projects, error=False)
            return projects.get_project_filenames()

    def _on_project_loaded(self, path):
        self.get_widget().update_active_project_path(path)

    def _on_project_closed(self):
        self.get_widget().update_active_project_path(None)

    # ---- Debugger related methods
    def _debugger_close_file(self, filename):
        debugger = self.get_plugin(Plugins.Debugger, error=False)
        if debugger is None:
            return True
        return debugger.can_close_file(filename)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)
# -----------------------------------------------------------------------------

# Third-party imports
import pytest

# Local imports
from spyder.config.manager import CONF
from spyder.plugins.appearance.plugin import Appearance
from spyder.widgets.config import SpyderConfigPage
from spyder.plugins.preferences.tests.conftest import (
    config_dialog, MainWindowMock)


@pytest.mark.parametrize(
    'config_dialog',
    [[MainWindowMock, [], [Appearance]]],
    indirect=True)
def test_change_ui_theme_and_color_scheme(config_dialog, mocker, qtbot):
    """Test that changing color scheme or UI theme works as expected."""
    # Patch methods whose calls we want to check
    mocker.patch.object(SpyderConfigPage, "prompt_restart_required")
    mocker.patch.object(CONF, "disable_notifications")

    # Get reference to Preferences dialog and widget page to interact with
    dlg = config_dialog
    widget = config_dialog.get_page()

    # List of color schemes
    names = widget.get_option('names')

    # Assert no restarts have been requested so far.
    assert SpyderConfigPage.prompt_restart_required.call_count == 0

    # Assert default UI theme is 'automatic' and interface is dark. The other
    # tests below depend on this.
    assert widget.get_option('ui_theme') == 'automatic'
    assert widget.is_dark_interface()

    # Change to another dark color scheme
    widget.schemes_combobox.setCurrentIndex(names.index('monokai'))
    dlg.apply_btn.click()
    assert SpyderConfigPage.prompt_restart_required.call_count == 0
    assert CONF.disable_notifications.call_count == 0

    # Change to a light color scheme
    widget.schemes_combobox.setCurrentIndex(names.index('pydev'))
    dlg.apply_btn.clicked.emit()
    assert SpyderConfigPage.prompt_restart_required.call_count == 1
    assert CONF.disable_notifications.call_count == 2

    # Change to the 'dark' ui theme
    widget.ui_combobox.setCurrentIndex(2)
    dlg.apply_btn.clicked.emit()
    assert SpyderConfigPage.prompt_restart_required.call_count == 1
    assert CONF.disable_notifications.call_count == 2

    # Change to the 'automatic' ui theme
    widget.ui_combobox.setCurrentIndex(0)
    dlg.apply_btn.clicked.emit()
    assert SpyderConfigPage.prompt_restart_required.call_count == 2
    assert CONF.disable_notifications.call_count == 4

    # Change to the 'light' ui theme
    widget.ui_combobox.setCurrentIndex(1)
    dlg.apply_btn.clicked.emit()
    assert SpyderConfigPage.prompt_restart_required.call_count == 3
    assert CONF.disable_notifications.call_count == 6

    # Change to another dark color scheme
    widget.schemes_combobox.setCurrentIndex(names.index('solarized/dark'))
    dlg.apply_btn.clicked.emit()
    assert SpyderConfigPage.prompt_restart_required.call_count == 4
    assert CONF.disable_notifications.call_count == 8

    # Change to the 'automatic' ui theme again
    widget.ui_combobox.setCurrentIndex(0)
    dlg.apply_btn.clicked.emit()
    assert SpyderConfigPage.prompt_restart_required.call_count == 4
    assert CONF.disable_notifications.call_count == 8



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# ----------------------------------------------------------------------------

# Standard library imports
import os
import os.path as osp
import random
import sys
import threading
import traceback

# Third-party imports
from jupyter_client.manager import KernelManager
from qtpy.QtCore import Qt
from qtpy.QtTest import QTest
from qtpy.QtWidgets import QApplication, QFileDialog, QLineEdit, QTabBar
# This is required to run our tests in VSCode or Spyder-unittest
from qtpy import QtWebEngineWidgets  # noqa
import psutil
import pytest

# Spyder imports
from spyder.api.plugin_registration.registry import PLUGIN_REGISTRY
from spyder.api.plugins import Plugins
from spyder.app import start
from spyder.config.base import get_home_dir, running_in_ci
from spyder.config.manager import CONF
from spyder.plugins.ipythonconsole.utils.kernelspec import SpyderKernelSpec
from spyder.plugins.projects.api import EmptyProject
from spyder.plugins.run.api import RunActions, StoredRunConfigurationExecutor
from spyder.plugins.toolbar.api import ApplicationToolbars
from spyder.utils import encoding
from spyder.utils.environ import (get_user_env, set_user_env,
                                  amend_user_shell_init)

# =============================================================================
# ---- Constants
# =============================================================================
# Location of this file
LOCATION = osp.realpath(osp.join(os.getcwd(), osp.dirname(__file__)))

# Time to wait until the IPython console is ready to receive input
# (in milliseconds)
SHELL_TIMEOUT = 40000 if os.name == 'nt' else 20000

# Need longer EVAL_TIMEOUT, because need to cythonize and C compile ".pyx" file
# before import and eval it
COMPILE_AND_EVAL_TIMEOUT = 30000

# Time to wait for the IPython console to evaluate something (in
# milliseconds)
EVAL_TIMEOUT = 3000

# Time to wait for the completion services to be up or give a response
COMPLETION_TIMEOUT = 30000


# =============================================================================
# ---- Auxiliary functions
# =============================================================================
def open_file_in_editor(main_window, fname, directory=None):
    """Open a file using the Editor and its open file dialog"""
    top_level_widgets = QApplication.topLevelWidgets()
    for w in top_level_widgets:
        if isinstance(w, QFileDialog):
            if directory is not None:
                w.setDirectory(directory)
            input_field = w.findChildren(QLineEdit)[0]
            input_field.setText(fname)
            QTest.keyClick(w, Qt.Key_Enter)


def reset_run_code(qtbot, shell, code_editor, nsb):
    """Reset state after a run code test"""
    qtbot.waitUntil(lambda: not shell._executing)
    with qtbot.waitSignal(shell.executed):
        shell.execute('%reset -f')
    qtbot.waitUntil(
        lambda: nsb.editor.source_model.rowCount() == 0, timeout=EVAL_TIMEOUT)
    code_editor.setFocus()
    qtbot.keyClick(code_editor, Qt.Key_Home, modifier=Qt.ControlModifier)


def start_new_kernel(startup_timeout=60, kernel_name='python', spykernel=False,
                     **kwargs):
    """Start a new kernel, and return its Manager and Client"""
    km = KernelManager(kernel_name=kernel_name)
    if spykernel:
        km._kernel_spec = SpyderKernelSpec()
    km.start_kernel(**kwargs)
    kc = km.client()
    kc.start_channels()
    try:
        kc.wait_for_ready(timeout=startup_timeout)
    except RuntimeError:
        kc.stop_channels()
        km.shutdown_kernel()
        raise

    return km, kc


def find_desired_tab_in_window(tab_name, window):
    all_tabbars = window.findChildren(QTabBar)
    for current_tabbar in all_tabbars:
        for tab_index in range(current_tabbar.count()):
            if current_tabbar.tabText(tab_index) == str(tab_name):
                return current_tabbar, tab_index
    return None, None


def read_asset_file(filename):
    """Read contents of an asset file."""
    return encoding.read(osp.join(LOCATION, filename))[0]


def create_project(tmpdir):
    """Create a simple project."""
    # Create project directory
    project = tmpdir.mkdir('test_project')
    project_path = str(project)

    # Create Spyder project
    spy_project = EmptyProject(project_path)
    CONF.set('project_explorer', 'current_project_path', project_path)

    # Add a file to the project
    p_file = project.join('file.py')
    p_file.write(read_asset_file('script_outline_1.py'))
    spy_project.set_recent_files([str(p_file)])


def create_complex_project(tmpdir):
    """Create a complex project."""
    # Create project directories
    project = tmpdir.mkdir('test_project')
    project_subdir = project.mkdir('subdir')
    project_sub_subdir = project_subdir.mkdir('sub_subdir')

    # Create directories out of the project
    out_of_project_1 = tmpdir.mkdir('out_of_project_1')
    out_of_project_2 = tmpdir.mkdir('out_of_project_2')
    out_of_project_1_subdir = out_of_project_1.mkdir('subdir')
    out_of_project_2_subdir = out_of_project_2.mkdir('subdir')

    project_path = str(project)
    spy_project = EmptyProject(project_path)
    CONF.set('project_explorer', 'current_project_path', project_path)

    # Add some files to project. This is necessary to test that we get
    # symbols for all these files.
    abs_filenames = []
    filenames_to_create = {
        project: ['file1.py', 'file2.py', 'file3.txt', '__init__.py'],
        project_subdir: ['a.py', '__init__.py'],
        project_sub_subdir: ['b.py', '__init__.py'],
        out_of_project_1: ['c.py'],
        out_of_project_2: ['d.py', '__init__.py'],
        out_of_project_1_subdir: ['e.py', '__init__.py'],
        out_of_project_2_subdir: ['f.py']
    }

    for path in filenames_to_create.keys():
        filenames = filenames_to_create[path]
        for filename in filenames:
            p_file = path.join(filename)
            abs_filenames.append(str(p_file))
            if osp.splitext(filename)[1] == '.py':
                if path == project_subdir:
                    code = read_asset_file('script_outline_2.py')
                elif path == project_sub_subdir:
                    code = read_asset_file('script_outline_3.py')
                else:
                    code = read_asset_file('script_outline_1.py')
                p_file.write(code)
            else:
                p_file.write("Hello world!")

    spy_project.set_recent_files(abs_filenames)


def create_namespace_project(tmpdir):
    """Create a project that contains a namespace package."""
    # Create project as example posted in:
    # https://github.com/spyder-ide/spyder/issues/16406#issuecomment-917992317
    project = tmpdir.mkdir('namespace-project')
    ns_package = project.mkdir('namespace-package')
    sub_package = ns_package.mkdir('sub-package')

    project_path = str(project)
    spy_project = EmptyProject(project_path)
    CONF.set('project_explorer', 'current_project_path', project_path)

    # Add some files to sub-package.
    abs_filenames = []
    filenames_to_create = {sub_package: ['module_1.py', '__init__.py']}

    for path in filenames_to_create.keys():
        filenames = filenames_to_create[path]
        for filename in filenames:
            p_file = path.join(filename)
            abs_filenames.append(str(p_file))

            # Use different files to be extra sure we're loading symbols in
            # each case.
            if filename == 'module.py':
                code = read_asset_file('script_outline_4.py')
            else:
                code = read_asset_file('script_outline_1.py')

            p_file.write(code)

    spy_project.set_recent_files(abs_filenames)


def preferences_dialog_helper(qtbot, main_window, section):
    """
    Open preferences dialog and select page with `section` (CONF_SECTION).
    """
    # Wait until the window is fully up
    shell = main_window.ipyconsole.get_current_shellwidget()
    qtbot.waitUntil(
        lambda: shell.spyder_kernel_ready and shell._prompt_html is not None,
        timeout=SHELL_TIMEOUT
    )

    main_window.show_preferences()
    preferences = main_window.preferences
    container = preferences.get_container()

    qtbot.waitUntil(lambda: container.dialog is not None, timeout=5000)
    dlg = container.dialog
    index = dlg.get_index_by_name(section)
    page = dlg.get_page(index)
    dlg.set_current_index(index)
    return dlg, index, page


def generate_run_parameters(mainwindow, filename, selected=None,
                            executor=None):
    """Generate run configuration parameters for a given filename."""
    file_uuid = mainwindow.editor.get_widget().id_per_file[filename]
    if executor is None:
        executor = mainwindow.ipyconsole.NAME

    file_run_params = StoredRunConfigurationExecutor(
        executor=executor,
        selected=selected,
    )

    return {file_uuid: [file_run_params]}


def get_random_dockable_plugin(main_window, exclude=None):
    """Get a random dockable plugin and give it focus."""
    plugins = main_window.get_dockable_plugins()
    for plugin_name, plugin in plugins:
        if exclude and plugin_name in exclude:
            plugins.remove((plugin_name, plugin))

    plugin = random.choice(plugins)[1]

    if not plugin.get_widget().toggle_view_action.isChecked():
        plugin.toggle_view(True)
        plugin._hide_after_test = True

    plugin.switch_to_plugin()
    plugin.get_widget().get_focus_widget().setFocus()
    return plugin


# =============================================================================
# ---- Pytest hooks
# =============================================================================
@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"
    setattr(item, "rep_" + rep.when, rep)


# =============================================================================
# ---- Fixtures
# =============================================================================
@pytest.fixture(scope="session", autouse=True)
def cleanup(request, qapp):
    """Cleanup the testing setup once we are finished."""

    def close_window():
        # Close last used mainwindow and QApplication if needed
        if hasattr(main_window, 'window') and main_window.window is not None:
            window = main_window.window
            main_window.window = None
            window.closing(close_immediately=True)
            window.close()
            window = None
            CONF.reset_to_defaults(notification=False)
            CONF.reset_manager()
            PLUGIN_REGISTRY.reset()

        if qapp.instance():
            for widget in qapp.allWidgets():
                try:
                    widget.close()
                except RuntimeError:
                    pass
            qapp.quit()

    request.addfinalizer(close_window)


@pytest.fixture
def main_window(request, tmpdir, qtbot):
    """Main Window fixture"""

    # Get original processEvents function in case the test that overrides it
    # fails
    super_processEvents = QApplication.processEvents

    # Don't show tours message
    CONF.set('tours', 'show_tour_message', False)

    # Tests assume inline backend
    CONF.set('ipython_console', 'pylab/backend', 'inline')

    # Test assume the plots are rendered in the console as png
    CONF.set('plots', 'mute_inline_plotting', False)
    CONF.set('ipython_console', 'pylab/inline/figure_format', "png")

    # Set exclamation mark to True
    CONF.set('debugger', 'pdb_use_exclamation_mark', True)

    # Check if we need to use introspection in a given test
    # (it's faster and less memory consuming not to use it!)
    use_introspection = request.node.get_closest_marker('use_introspection')

    if use_introspection:
        CONF.set('completions', ('enabled_providers', 'lsp'), True)
        CONF.set('completions', ('enabled_providers', 'fallback'), True)
        CONF.set('completions', ('enabled_providers', 'snippets'), True)
    else:
        CONF.set('completions', ('enabled_providers', 'lsp'), False)
        CONF.set('completions', ('enabled_providers', 'fallback'), False)
        CONF.set('completions', ('enabled_providers', 'snippets'), False)

    # Only use single_instance mode for tests that require it
    single_instance = request.node.get_closest_marker('single_instance')

    if single_instance:
        CONF.set('main', 'single_instance', True)
    else:
        CONF.set('main', 'single_instance', False)

    # Check if we need to load a simple project to the interface
    preload_project = request.node.get_closest_marker('preload_project')
    if preload_project:
        create_project(tmpdir)
    else:
        CONF.set('project_explorer', 'current_project_path', None)

    # Check if we need to preload a complex project in a give test
    preload_complex_project = request.node.get_closest_marker(
        'preload_complex_project')
    if preload_complex_project:
        CONF.set('editor', 'show_class_func_dropdown', True)
        create_complex_project(tmpdir)
    else:
        CONF.set('editor', 'show_class_func_dropdown', False)
        if not preload_project:
            CONF.set('project_explorer', 'current_project_path', None)

    # Check if we need to preload a project with a namespace package
    preload_namespace_project = request.node.get_closest_marker(
        'preload_namespace_project')
    if preload_namespace_project:
        create_namespace_project(tmpdir)
    else:
        if not (preload_project or preload_complex_project):
            CONF.set('project_explorer', 'current_project_path', None)

    # Get config values passed in parametrize and apply them
    try:
        param = request.param
        if isinstance(param, dict) and 'spy_config' in param:
            CONF.set(*param['spy_config'])
    except AttributeError:
        # Not all tests that use this fixture define request.param
        pass

    QApplication.processEvents()

    if not hasattr(main_window, 'window') or main_window.window is None:
        # Start the window
        window = start.main()
        main_window.window = window

    else:
        window = main_window.window

        if not request.node.get_closest_marker('no_new_console'):
            # Create a new console to ensure new config is loaded
            # even if the same mainwindow instance is reused
            window.ipyconsole.create_new_client(give_focus=True)

    # Add a handle to the "Debug file" button to access it quickly because
    # it's used a lot.
    toolbar = window.get_plugin(Plugins.Toolbar)
    debug_toolbar = toolbar.get_application_toolbar(ApplicationToolbars.Debug)
    debug_action = window.run.get_action(
        "run file in debugger")
    debug_button = debug_toolbar.widgetForAction(debug_action)
    window.debug_button = debug_button

    # Add a handle to the run buttons to access it quickly because they are
    # used a lot.
    run_toolbar = toolbar.get_application_toolbar(ApplicationToolbars.Run)
    run_action = window.run.get_action(RunActions.Run)
    run_button = run_toolbar.widgetForAction(run_action)
    window.run_button = run_button

    run_cell_action = window.run.get_action('run cell')
    run_cell_button = run_toolbar.widgetForAction(run_cell_action)
    window.run_cell_button = run_cell_button

    run_cell_and_advance_action = window.run.get_action('run cell and advance')
    run_cell_and_advance_button = run_toolbar.widgetForAction(
        run_cell_and_advance_action)
    window.run_cell_and_advance_button = run_cell_and_advance_button

    run_selection_action = window.run.get_action('run selection and advance')
    run_selection_button = run_toolbar.widgetForAction(run_selection_action)
    window.run_selection_button = run_selection_button

    QApplication.processEvents()

    if os.name != 'nt':
        # _DummyThread are created if current_thread() is called from them.
        # They will always leak (From python doc) so we ignore them.
        init_threads = [
            repr(thread) for thread in threading.enumerate()
            if not isinstance(thread, threading._DummyThread)]
        proc = psutil.Process()
        init_files = [repr(f) for f in proc.open_files()]
        init_subprocesses = [repr(f) for f in proc.children()]

    yield window

    # Remap original QApplication.processEvents function
    QApplication.processEvents = super_processEvents

    # Print shell content if failed
    if request.node.rep_setup.passed:
        if request.node.rep_call.failed:
            # Print content of shellwidget and close window
            print(window.ipyconsole.get_current_shellwidget(
                )._control.toPlainText())
            # Print info page content is not blank
            console = window.ipyconsole
            client = console.get_current_client()
            if client.info_page != client.blank_page:
                print('info_page')
                print(client.info_page)
            main_window.window = None
            window.closing(close_immediately=True)
            window.close()
            window = None
            CONF.reset_to_defaults(notification=False)
            CONF.reset_manager()
            PLUGIN_REGISTRY.reset()

        else:
            # Try to close used mainwindow directly on fixture
            # after running test that uses the fixture
            # Currently 'test_out_runfile_runcell' is the last tests so
            # in order to prevent errors finalizing the test suit such test has
            # this marker.
            # Also, try to decrease chances of freezes/timeouts from tests that
            # are known to have leaks by also closing the main window for them.
            known_leak = request.node.get_closest_marker(
                'known_leak')
            close_main_window = request.node.get_closest_marker(
                'close_main_window')
            if close_main_window or known_leak:
                main_window.window = None
                window.closing(close_immediately=True)
                window.close()
                window = None
                CONF.reset_to_defaults(notification=False)
                CONF.reset_manager()
                PLUGIN_REGISTRY.reset()
            else:
                try:
                    # Close or hide everything we can think of
                    window.switcher.hide()
                    window.switcher.on_close()

                    # Close editor related elements
                    window.editor.close_all_files()

                    # Force close all files
                    editor_widget = window.editor.get_widget()
                    while editor_widget.editorstacks[0].close_file(force=True):
                        pass
                    for editorwindow in editor_widget.editorwindows:
                        editorwindow.close()

                    window.projects.close_project()

                    if window.console.error_dialog:
                        window.console.close_error_dialog()

                    # Reset cwd
                    window.explorer.chdir(get_home_dir())

                    # Restore default Spyder Python Path
                    CONF.set(
                        'pythonpath_manager', 'spyder_pythonpath',
                        CONF.get_default('pythonpath_manager',
                                         'spyder_pythonpath')
                    )

                    # Restore run configurations
                    CONF.set('run', 'configurations', [])

                    # Close consoles
                    (window.ipyconsole.get_widget()
                        .create_new_client_if_empty) = False
                    window.ipyconsole.restart()
                except Exception:
                    main_window.window = None
                    window.closing(close_immediately=True)
                    window.close()
                    window = None
                    CONF.reset_to_defaults(notification=False)
                    CONF.reset_manager()
                    PLUGIN_REGISTRY.reset()
                    return

                if os.name == 'nt':
                    # Do not test leaks on windows
                    return

                def show_diff(init_list, now_list, name):
                    sys.stderr.write(f"Extra {name} before test:\n")
                    for item in init_list:
                        if item in now_list:
                            now_list.remove(item)
                        else:
                            sys.stderr.write(item + "\n")
                    sys.stderr.write(f"Extra {name} after test:\n")
                    for item in now_list:
                        sys.stderr.write(item + "\n")

                # The test is not allowed to open new files or threads.
                try:
                    def threads_condition():
                        threads = [
                            thread for thread in threading.enumerate()
                            if not isinstance(thread, threading._DummyThread)]
                        return (len(init_threads) >= len(threads))

                    qtbot.waitUntil(threads_condition, timeout=SHELL_TIMEOUT)
                except Exception:
                    now_threads = [
                        thread for thread in threading.enumerate()
                        if not isinstance(thread, threading._DummyThread)]
                    threads = [repr(t) for t in now_threads]
                    show_diff(init_threads, threads, "thread")
                    sys.stderr.write("Running Threads stacks:\n")
                    now_thread_ids = [t.ident for t in now_threads]
                    for thread_id, frame in sys._current_frames().items():
                        if thread_id in now_thread_ids:
                            sys.stderr.write(
                                "\nThread " + str(threads) + ":\n")
                            traceback.print_stack(frame)
                    main_window.window = None
                    window.closing(close_immediately=True)
                    window.close()
                    window = None
                    CONF.reset_to_defaults(notification=False)
                    CONF.reset_manager()
                    PLUGIN_REGISTRY.reset()
                    raise

                try:
                    qtbot.waitUntil(lambda: (
                        len(init_subprocesses) >= len(proc.children())),
                        timeout=SHELL_TIMEOUT)
                except Exception:
                    subprocesses = [repr(f) for f in proc.children()]
                    show_diff(init_subprocesses, subprocesses, "processes")
                    main_window.window = None
                    window.closing(close_immediately=True)
                    window.close()
                    window = None
                    CONF.reset_to_defaults(notification=False)
                    CONF.reset_manager()
                    PLUGIN_REGISTRY.reset()
                    raise

                try:
                    files = [
                        repr(f) for f in proc.open_files()
                        if 'QtWebEngine' not in repr(f)
                    ]
                    qtbot.waitUntil(
                        lambda: (len(init_files) >= len(files)),
                        timeout=SHELL_TIMEOUT)
                except Exception:
                    show_diff(init_files, files, "files")
                    main_window.window = None
                    window.closing(close_immediately=True)
                    window.close()
                    window = None
                    CONF.reset_to_defaults(notification=False)
                    CONF.reset_manager()
                    PLUGIN_REGISTRY.reset()
                    raise


@pytest.fixture
def restore_user_env():
    """Set user environment variables and restore upon test exit"""
    if not running_in_ci():
        pytest.skip("Skipped because not in CI.")

    if os.name == "nt":
        orig_env = get_user_env()

    yield

    if os.name == "nt":
        set_user_env(orig_env)
    else:
        amend_user_shell_init(restore=True)
